#
# Excel : process the data in .xlsx file
#
# Example:
# ::
#     import office
#
#     # open file datafile.xlsx, return workbook object
#     workbook = office.open_file("datafile.xlsx")
#
#     # list worksheets
#     sheet_names = workbook.sheetnames
#     print(sheet_names)
#
#     # get a sheet
#     sheet = workbook['Sales']
#     print(sheet.name)
#
#     # get a cell
#     cell = sheet['B2']
#     print(cell.value)
#
#     # write a cell
#     cell.value = 'New City'
#
#     # save
#     workbook.save("newfile.xlsx")
#
#     # close
#     workbook.close()
#
#

import os
import time
import inspect
from .base import BaseFile
from .util import Util
from .win32 import Win32
import re
import openpyxl

from openpyxl.styles.numbers import FORMAT_GENERAL, FORMAT_TEXT, FORMAT_NUMBER
from openpyxl.styles.numbers import FORMAT_PERCENTAGE, FORMAT_DATE_DATETIME

try:
    import xlwings
except ImportError:
    xlwings = None


try:
    import pandas
    from pandas.io.formats.excel import ExcelFormatter
    from pandas import ExcelWriter
except ImportError:
    pandas = None
    ExcelFormatter = None
    ExcelWriter = None


class XlsConsts:
    # https://docs.microsoft.com/en-us/office/vba/api/word.wdsaveformat
    wdFormatDocument = 0   # Microsoft Office Word 97 - 2003 binary file format.
    wdFormatTemplate = 1  # Word template format.
    wdFormatText = 2  # Microsoft Windows text format.
    wdFormatRTF = 6  # Rich text format (RTF).
    wdFormatHTML = 8  # Standard HTML format.
    wdFormatWebArchive = 9  # Web archive format.
    wdFormatFilteredHTML = 10  # Filtered HTML format.
    wdFormatXMLTemplate = 14  # XML template format.
    wdFormatPDF = 17  # PDF format.


class XlsUtil:

    @staticmethod
    def is_formula(value):
        """ whether value is a formula of Excel"""
        if isinstance(value, str) and value.startswith('='):
            return True
        return False

    @staticmethod
    def is_hex_char(c: str):
        """ whether char c is a hexadecimal char """
        n = ord(c)
        if (ord('0') <= n <= ord('9')) or (ord('A') <= n <= ord('F')):
            return True
        return False

    @staticmethod
    def is_hex_str(s: str):
        """ whether the string is a hexadecimal string """
        if not isinstance(s, str):
            return False

        s = s[1:] if s.startswith('#') else s
        s = s.upper()
        for c in s:
            if not XlsUtil.is_hex_char(c):
                return False
        return True

    @staticmethod
    def to_hex_str(n: int):
        """ convert specified number to hexadecimal string """
        s = hex(n)
        if s.startswith('0x'):
            s = s[2:]
        if len(s) % 2 == 1:
            s += '0'
        return s.upper()

    @staticmethod
    def is_int_str(s: str):
        """ whether the string is an integer string"""
        if not isinstance(s, str):
            return False

        if s == '':
            return False

        for c in s:
            if not ('0' <= c <= '9'):
                return False
        return True

    @staticmethod
    def is_color(value):
        """ whether the value is a color string """
        if isinstance(value, str) and XlsUtil.is_hex_str(value) and len(value) >= 3:
            return True
        return False

    @staticmethod
    def to_color_str(value):
        """  convert the value to a color string"""
        if isinstance(value, str) and XlsUtil.is_hex_str(value):
            if value.startswith('#'):
                value = value[1:]
                if len(value) == 3:
                    value = value[0] + value[0] + value[1] + value[1] + value[2] + value[2]
            else:
                value = value.upper()
        return value

    @staticmethod
    def is_number(value):
        """ whether the value is number """
        if isinstance(value, int) or isinstance(value, float):
            if not isinstance(value, bool):
                return True
        return False

    @staticmethod
    def rgb(r: int, g: int, b: int) -> str:
        """ return rgb color string """
        return XlsUtil.to_hex_str(r) + XlsUtil.to_hex_str(g) + XlsUtil.to_hex_str(b)

    @staticmethod
    def convert_number_format(value):
        if value is None or value == '':
            return FORMAT_GENERAL

        if value == 'text' or value == '%s':
            return FORMAT_TEXT

        if value == 'number':
            return FORMAT_NUMBER

        if value == 'percentage' or value == 'percent':
            return FORMAT_PERCENTAGE

        if value == 'date' or value == 'datetime':
            return FORMAT_DATE_DATETIME

        if re.search('%[YyHhMmS]', value) is not None:
            value = value.replace('%Y', 'yyyy')
            value = value.replace('%y', 'yy')
            value = value.replace('%m', 'mm')
            value = value.replace('%d', 'dd')
            value = value.replace('%H', 'hh')
            value = value.replace('%M', 'mm')
            value = value.replace('%S', 'ss')
            return value

        if re.search('%(.*)f', value) is not None:
            value = XlsUtil.to_excel_number_format(value)

        return value

    @staticmethod
    def to_excel_number_format(format_str: str):
        """
        convert printf() float format string to Excel number format

        :param format_str:  float format string in printf(), such as '%8.3f'

        :return: return Excel number format, example:  '%8.3f'  ->  #,###.000
        """
        value = format_str

        # find substring like '%??f'
        m = re.match('%(.*)f', value)
        if m is None:
            return '0'

        # get the '%??f' substring
        span = m.span()
        s = value[span[0] + 1: span[1] - 1]

        # split by char '.'
        arr = s.split('.')
        suffix = value[span[1]:].strip()

        if len(arr) == 1 and arr[0] != '':
            length = int(arr[0]) if XlsUtil.is_int_str(arr[0]) else 1
            length = -length if length < 0 else length

            if suffix == ',':
                return '#,###'
            else:
                return '#' * length

        elif len(arr) == 2:
            length = int(arr[0]) if XlsUtil.is_int_str(arr[0]) else 1
            length = -length if length < 0 else length

            digit = int(arr[1]) if XlsUtil.is_int_str(arr[1]) else 0
            length = digit + 1 if length < digit else length

            if suffix == ',':
                return '#,###' + '.' + ('0' * digit)
            else:
                return ('#' * (length - digit)) + '.' + ('0' * digit)
        else:
            return '0'

    @staticmethod
    def col_str_to_int(col_str: str):
        """
        convert Excel column string to integer, such as 'A' -> 1,  'D' -> 4

        :param col_str:  column string, such as 'A', 'BZ'

        :return: return int.  example: 'A' -> 1,  'D' -> 4
        """
        col_str = col_str.upper()
        num = 0
        for i in range(0, len(col_str)):
            num *= 26
            c = col_str[i]
            if ord('A') <= ord(c) <= ord('Z'):
                num += ord(c) - ord('A') + 1
            else:
                return -1
        return num

    @staticmethod
    def col_int_to_str(col_num: int):
        """
        convert integer to Excel column string. such as 1 -> 'A',  4 -> 'D'

        :param col_num: column number ( First column 'A' is 1)

        :return: return string, example: 1 -> 'A',  4 -> 'D'
        """
        result = ''
        while True:
            if col_num > 26:
                col_num, r = divmod(col_num - 1, 26)
                result = chr(r + ord('A')) + result
            else:
                return chr(col_num + ord('A') - 1) + result

    @staticmethod
    def is_list_like(obj):
        """
        Whether the object is like a list

        :param obj: the object

        :return:
        """
        if isinstance(obj, list) or isinstance(obj, tuple):
            return True

        if (not isinstance(obj, str)) and hasattr(obj, '__getitem__') \
                and hasattr(obj, '__len__') and (not isinstance(obj, dict)):
            return True

        return False


class Cell:
    def __init__(self, sheet, obj):
        self.sheet: WorkSheet = sheet
        self.cell_obj = None
        self._load(obj)

    def _load(self, obj):
        """
        读取单元格配置

        :param obj: Address string or tuple(row, col) or cell object
        :return:
        """
        if isinstance(obj, str):
            c = self.sheet[obj]
            if isinstance(c, Cell):
                self.cell_obj = c.cell_obj
            elif isinstance(c, Range):
                self.cell_obj = c.cell1.cell_obj
        elif Util.is_list_like(obj) and len(obj) == 2 and \
                isinstance(obj[0], int) and isinstance(obj[1], int):
            self.cell_obj = self.sheet.cell(obj[0], obj[1]).cell_obj
        else:
            self.cell_obj = obj

    @property
    def value(self):
        """ Cell的值 """
        return self.cell_obj.value

    @value.setter
    def value(self, val):
        if type(val).__name__ == 'DataFrame':
            self.sheet.write_dataframe(self.address, val)
        else:
            self.cell_obj.value = val

    @property
    def row(self):
        """ Cell的行号, 第一行为 1 """
        return self.cell_obj.row

    @property
    def column(self):
        """ Cell的列号, A列为 1 """
        return self.cell_obj.column

    @property
    def address(self):
        """ Cell的地址，例如 'A3' """
        return self.to_str(absolute=False)

    def to_str(self, absolute=False):
        """
        转为字符串
        convert to string
        :param absolute: (optional) whether display absolute mark '$'
        :return:  return a string
        """
        result = ''
        if self.column >= 0:
            if absolute:
                result += '$'
            result += XlsUtil.col_int_to_str(self.column)

        if self.row >= 0:
            if absolute:
                result += '$'
            result += str(self.row)

        return result

    def move(self, delta_row: int, delta_col: int):
        """
        将单元格移动到一个新的位置

        :param delta_row:   the rows to move. negative number means move left
        :param delta_col:  the columns to move. negative number means move up
        :return:  self
        """
        new_row = self.row + delta_row
        new_col = self.column + delta_col
        if new_row <= 0 or new_col <= 0:
            raise ValueError("row or column is little than 0")
        self.cell_obj = self.sheet.cell(new_row, new_col).cell_obj
        return self

    def expand(self, right=0, down=0):
        """
        水平或垂直方向 扩展区域

        :param right:   水平方向扩展的列数. 正数表示向右，负数表示向左，0表示向右扩展到非空单元格，None表示不扩展。
        :param down:   垂直方向扩展的行数. 正数表示向下，负数表示向上，0表示向下扩展到非空单元格，None表示不扩展。
        :return:  Range
        """
        return Range(self.sheet, self).expand(right, down)  # TODO

    def equals(self, cell):
        """ 判断本单元格是否等于指定单元格cell"""
        if self.sheet.sheet_obj == self.sheet.sheet_obj:
            if self.row == cell.row and self.column == cell.column:
                return True

    @property
    def column_width(self):
        """ 列宽 """
        if self.sheet.book.is_xlwings:
            return self.cell_obj.column_width
        else:
            col = XlsUtil.col_int_to_str(self.column)
            return self.sheet.sheet_obj.column_dimensions[col].width

    @column_width.setter
    def column_width(self, value):
        if self.sheet.book.is_xlwings:
            self.cell_obj.column_width = value
        else:
            col = XlsUtil.col_int_to_str(self.column)
            self.sheet.sheet_obj.column_dimensions[col].width = value

    @property
    def row_height(self):
        """ 行高 """
        if self.sheet.book.is_xlwings:
            return self.cell_obj.row_height
        else:
            return self.sheet.sheet_obj.row_dimensions[self.row].height

    @row_height.setter
    def row_height(self, value):
        if self.sheet.book.is_xlwings:
            self.cell_obj.row_height = value
        else:
            self.sheet.sheet_obj.row_dimensions[self.row].height = value

    @property
    def number_format(self):
        """ 数据格式 """
        return self.cell_obj.number_format

    @number_format.setter
    def number_format(self, value):
        self.cell_obj.number_format = XlsUtil.convert_number_format(value)

    def __repr__(self):
        return '<Cell %s>' % repr(self.address)


class RangeList(list):
    pass


class Range:
    def __init__(self, sheet, obj, cell2=None):
        self.sheet: WorkSheet = sheet
        self.cell1: Cell = None
        self.cell2: Cell = None
        self.range_obj = None
        self._load(obj, cell2)

    def _load(self, obj, cell2=None):
        """
        设置Range区域

        :param obj:  可以是一个地址字符串，可以是Cell对象，可以是 xlwings的Range对象， openpyxl的tuple
        :param cell2:  可以是Cell对象，可以是 xlwings的Range对象， openpyxl的tuple
        :return: None
        """
        if isinstance(obj, str):
            obj = self.sheet[obj]

        if self.sheet.book.is_xlwings:
            if type(obj).__name__ == 'Range':
                self.range_obj = obj
                self.cell1 = self.sheet.cell(self.range_obj.row, self.range_obj.column)
                self.cell2 = self.sheet.cell(self.range_obj.row + len(self.range_obj.rows) - 1,
                                             self.range_obj.column + len(self.range_obj.columns) - 1)
            elif isinstance(obj, Cell):
                self.cell1 = obj
                self.cell2 = cell2 if isinstance(cell2, Cell) else self.cell1
                self.rect()
                self.range_obj = self.sheet.sheet_obj[self.cell1.address + ':' + self.cell2.address]
            elif Util.is_list_like(obj) and len(obj) == 2 and Util.is_list_like(cell2) and len(cell2) == 2:
                self.cell1 = self.sheet.cell(obj[0], obj[1])
                self.cell2 = self.sheet.cell(cell2[0], cell2[1])
                self.rect()
                self.range_obj = self.sheet.sheet_obj[self.cell1.address + ':' + self.cell2.address]
            else:
                raise ValueError("%s is not a Range" % repr(type(obj)))
        else:
            if isinstance(obj, tuple):
                row = obj[0]
                self.cell1 = Cell(self, row[0])

                row = obj[len(obj) - 1]
                self.cell2 = Cell(self, row[len(row) - 1])
                self.rect()
            elif isinstance(obj, Cell):
                self.cell1 = obj
                self.cell2 = cell2 if isinstance(cell2, Cell) else self.cell1
                self.rect()
                self.range_obj = None
            else:
                raise ValueError("%s is not a Range" % repr(type(obj)))

    def rect(self):
        """ return start_row, start_col, end_row, end_col of the rectangle of the Range """
        start_row = self.cell1.row
        start_col = self.cell1.column
        end_row = self.cell2.row
        end_col = self.cell2.column
        if end_row < start_row:
            start_row, end_row = end_row, start_row
        if end_col < start_col:
            start_col, end_col = end_col, start_col
        if self.cell1.row != start_row or self.cell1.column != start_col:
            self.cell1 = self.sheet.cell(start_row, start_col)
        if self.cell2.row != end_row or self.cell2.column != end_col:
            self.cell2 = self.sheet.cell(end_row, end_col)
        return start_row, start_col, end_row, end_col

    @property
    def value(self):
        """ 区域的值， 返回一个list"""
        if self.sheet.book.is_xlwings:
            return self.range_obj.value

        else:
            result = []
            for r in range(self.cell1.row, self.cell2.row + 1):
                data = []
                for c in range(self.cell1.column, self.cell2.column + 1):
                    data.append(self.sheet.cell(r, c).value)
                result.append(data)
            return result

    @value.setter
    def value(self, val):
        if self.sheet.book.is_xlwings:
            self.range_obj.value = val

        else:
            if not Util.is_list_like(val):
                val = [[val]]

            if Util.is_table_list(val):
                start_row, start_col, end_row, end_col = self.rect()
                for r in range(0, end_row - start_row + 1):
                    for c in range(0, end_col - start_col + 1):
                        row_val = val[r] if len(val) > r else []
                        if len(row_val) > c:
                            self.sheet.cell(start_row + r, start_col + c).value = row_val[c]

    def clear(self, value=None):
        """
        清空区域， 填入默认值 default value

        :param value: (optional) value which is filled into cells
        :return: self
        """
        # scan the area, set each cell's value
        for row in range(self.cell1.row, self.cell2.row + 1):
            for col in range(self.cell1.column, self.cell2.column + 1):
                self.sheet.cell(row, col).value = value
        return self

    @property
    def address(self):
        """ Range的地址，例如 'A3:B6' """
        return self.cell1.address + ':' + self.cell2.address

    @property
    def rows(self):
        """ Return a list the rows in the range """
        rs = RangeList()
        start_row, start_col, end_row, end_col = self.rect()
        for row in range(start_row, end_row + 1):
            cell1 = self.sheet.cell(row, start_col)
            cell2 = self.sheet.cell(row, end_col)
            rs.append(Range(self.sheet, cell1, cell2))
        return rs

    @property
    def columns(self):
        """ Return a list the columns in the range """
        rs = RangeList()
        start_row, start_col, end_row, end_col = self.rect()
        for col in range(start_col, end_col + 1):
            cell1 = self.sheet.cell(start_row, col)
            cell2 = self.sheet.cell(end_row, col)
            rs.append(Range(self.sheet, cell1, cell2))
        return rs

    def expand(self, right=0, down=0):
        """
        水平或垂直方向 扩展区域

        :param right:   水平方向扩展区域的列数. 正数表示向右，负数表示向左，0表示向右扩展到非空单元格，None表示不扩展。
                        the rows to expand rightward. negative number is not allowed
        :param down:   垂直方向扩展区域的行数. 正数表示向下，负数表示向上，0表示向下扩展到非空单元格，None表示不扩展。
                       the columns to downward. negative number is not allowed
        :return:  self
        """
        row = self.cell2.row
        col = self.cell2.column

        # expand horizontal
        if isinstance(right, int):
            if right == 0:
                cell = self.sheet.cell(row, col + 1)
                while cell.value is not None:
                    col += 1
                    cell = self.sheet.cell(row, col + 1)
            else:
                col += right
                if col < 1:
                    col = 1

        # expand vertical
        if isinstance(down, int):
            if down == 0:
                cell = self.sheet.cell(row + 1, col)
                while cell.value is not None:
                    row += 1
                    cell = self.sheet.cell(row + 1, col)
            else:
                row += down
                if row < 1:
                    row = 1

        self.cell2 = self.sheet.cell(row, col)
        return self

    def to_list(self):
        """ 读取区域内各单元格的值，返回 list """
        return self.value

    def to_dataframe(self):
        """ 读取区域内各单元格的值，返回DataFrame对象 """
        return self.sheet.read_dataframe(self)

    def merge_cells(self):
        """ 合并区域内所有单元格 """
        if self.sheet.book.is_xlwings:
            self.range_obj.merge_cells()
        else:
            self.sheet.sheet_obj.merge_cells(self.address)

    def is_cell(self):
        """ 判断本区域是否只有一个单元格 """
        if self.cell1.row == self.cell2.row and self.cell1.column == self.cell2.column:
            return True
        return False

    def __repr__(self):
        return '<Range %s>' % repr(self.address)


class WorkSheet:
    def __init__(self, book, sheet_obj):
        self.book: Excel = book
        self.sheet_obj = sheet_obj

    @property
    def max_column(self):
        """ the max columns that worksheet used  """
        if self.book.is_xlwings:
            all_cells = self.sheet_obj.cells
            count = all_cells.columns.count
            max_c = 1
            for row in range(1, 200):
                r = 100 + (row - 100) * 2 if row > 100 else row
                cur_c = self.range((r, count)).range_obj.end('left').column
                max_c = max(max_c, cur_c)
            return max_c + 1
        else:
            return self.sheet_obj.max_column

    @property
    def max_row(self):
        """ the max row that worksheet used """
        if self.book.is_xlwings:
            all_cells = self.sheet_obj.cells
            count = all_cells.rows.count
            max_r = 1
            for col in range(1, 200):
                c = 100 + (col - 100) * 2 if col > 100 else col
                cur_r = self.range((count, c)).range_obj.end('up').row
                max_r = max(max_r, cur_r)
            return max_r + 1
        else:
            return self.sheet_obj.max_row

    @property
    def title(self):
        """ the title of the worksheet ( alias of name) """
        if self.book.is_xlwings:
            return self.sheet_obj.name
        else:
            return self.sheet_obj.title

    @title.setter
    def title(self, value):
        if self.book.is_xlwings:
            self.sheet_obj.name = value
        else:
            self.sheet_obj.title = value

    @property
    def name(self):
        """ the name of the sheet """
        return self.title

    @name.setter
    def name(self, value):
        self.title = value

    def cell(self, row: int, column: int):
        """
        Get a cell of specified row, column

        :Chinese: 根据行号、列号，返回一个Cell对象

        :param row:   row number(int), the first row is 1
        :param column: column number(int), the first column is 1
        :return: return a Cell objecdt
        """
        if self.book.is_xlwings:
            return Cell(self, self.sheet_obj.range((row, column)))
        else:
            return Cell(self, self.sheet_obj.cell(row, column))

    def range(self, cell1, cell2=None):
        """
        get a range

        :param cell1: left-top cell
        :param cell2: (optinal)right-bottom cell
        :return: return a Range object
        """
        cell2 = cell1 if cell2 is None else cell2
        return Range(self, cell1, cell2)

    def __getitem__(self, item) -> [Cell, Range]:
        """ get a cell or a range """

        if isinstance(item, str):
            obj = self.sheet_obj[item]
            if self.book.is_xlwings:
                if len(obj.rows) == 1 and len(obj.columns) == 1:
                    return Cell(self, obj)
                else:
                    return Range(self, obj)
            else:
                if isinstance(obj, tuple):
                    return Range(self, obj)
                else:
                    return Cell(self, obj)

    @property
    def index(self):
        """ return the index of the sheet in the parent workbook """
        if self.book.is_xlwings:
            return self.sheet_obj.index
        else:
            return self.title in self.sheet_obj.sheetnames

    def find(self, value, value2=None, regex=False):
        """
        Find a cell

        :param value1:   first value
        :param value2:  (optional) second value
        :param regex:   (optional) whether use regex

        :return: return Cell object if found. return None if not found.
        """
        if value2 is not None:
            c1 = self.find(value, regex=regex)
            c2 = self.find(value2, regex=regex)
            if c1 and c2:
                return Cell(self, (c1.row, c2.column))

        max_row = self.max_row
        max_column = self.max_column
        for row in range(1, max_row):
            for col in range(1, max_column):
                val = self.cell(row, col).value
                if not regex:
                    if value == val:
                        return Cell(self, (row, col))
                else:
                    if isinstance(val, str) and re.search(value, str(val)) is not None:
                        return Cell(self, (row, col))

        return None

    def _address_to_range(self, address):
        """ convert address to Range """
        if isinstance(address, Range):
            return address
        else:
            c = self[address]
            if isinstance(c, Cell):
                return Range(c.sheet, c, c)
            else:
                return c

    def read_list(self, address):
        """
        read list from specified address

        :param address: the address, such as: 'A2:C4'
        :return: return a list
        """
        rng = self._address_to_range(address)
        return rng.to_list()

    def write_list(self, address, list_data, vertical=False):
        """
        write a list to the specified address

        :param address:   the address, such as: 'A2:C4'
        :param list_data: list of data
        :param vertical:  (optional)whether write vertically
        :return:
        """
        rng = self._address_to_range(address)
        start_row, start_col, end_row, end_col = rng.rect()

        for r, row_data in enumerate(list_data):
            if Util.is_list_like(row_data):
                for c, val in enumerate(row_data):
                    if not vertical:
                        row = start_row + r
                        col = start_col + c
                    else:
                        row = start_row + c
                        col = start_col + r
                    if (row <= end_row or end_row == 0) and (col <= end_col or end_col == 0):
                        self.cell(row, col).value = val
            else:
                if not vertical:
                    row = start_row
                    col = start_col + r
                else:
                    row = start_row + r
                    col = start_col
                self.cell(row, col).value = row_data

    def read_dataframe(self, address):
        """
         read DataFrame from specified address

        :param address:  the address, such as: 'A2:C4'
        :return: DataFrame
        """
        data = self.read_list(address)
        if len(data) > 0:
            return pandas.DataFrame(data[1:], columns=data[0])
        else:
            return pandas.DataFrame()

    def write_dataframe(self, address, df, header=True, vertical=False, **kwargs):
        """
        write pandas DataFrame

        :param address:     the address
        :param df:          DataFrame object
        :param header:      whether write header of the DataFrame object
        :param vertical:    vertical placed
        :return: return a tuple of (rows, cols) which are the count of rows and cols that be written.
        """
        if not self.book.is_xlwings and not vertical:
            self._openpyxl_write_dataframe(address, df, header, **kwargs)
            return

        rng = self._address_to_range(address)
        start_row, start_col, end_row, end_col = rng.rect()

        h = 1 if header else 0

        if end_row == start_row and end_col == start_col:
            if vertical:
                end_row = start_row + len(df.columns) - 1 if len(df.columns) > 0 else start_row
            else:
                end_row = start_row + len(df) - 1 + h if len(df) > 0 else start_row

        if end_col == start_col and end_col == start_col:
            if vertical:
                end_col = start_col + len(df) - 1 + h if len(df) > 0 else end_col
            else:
                end_col = start_col + len(df.columns) - 1 if len(df.columns) > 0 else end_col

        count = 0
        # write dataframe columns
        if header:
            if vertical:
                r = Range(self, (start_row, start_col + count), (end_row, start_col + count))
            else:
                r = Range(self, (start_row + count, start_col), (start_row + count, end_col))
            self.write_list(r, list(df.columns), vertical)
            count += 1

        # write dataframe data of each row
        for i in range(len(df)):
            if vertical:
                r = Range(self, (start_row, start_col + count), (end_row, start_col + count))
            else:
                r = Range(self, (start_row + count, start_col), (start_row + count, end_col))
            self.write_list(r, list(df.iloc[i]), vertical)
            count += 1

    def _openpyxl_write_dataframe(self, address, df, header=True, **kwargs):
        rng = self[address]
        if isinstance(rng, Cell):
            start_cell = rng
        else:
            start_cell = rng.cell1

        sheet_name = self.name
        writer = ExcelWriter(self.book.filename, engine='openpyxl')
        writer.book = self.sheet_obj
        writer._get_sheet = dict((ws.title, ws) for ws in self.sheet_obj.worksheets)

        index = kwargs.pop('index') if 'index' in kwargs else False
        header = kwargs.pop('header') if 'header' in kwargs else header

        ExcelFormatter.header_style = None

        self.book.save()
        df.to_excel(writer, sheet_name=sheet_name, index=index, header=header,
                    startrow=start_cell.row - 1, startcol=start_cell.column - 1, **kwargs)

        writer.save()
        writer.close()

        delta = 1 if header else 0
        return len(df) + delta, len(df.columns)

    def protect(self, password: str):
        """ Protect the Worksheet with password """
        if self.book.is_xlwings:
            self.sheet_obj.api.Protect(password)
        else:
            self.sheet_obj.protection.sheet = True
            self.sheet_obj.protection.password = password
            self.sheet_obj.protection.enable()

    def unprotect(self, password: str):
        """ Unprotect the WorkSheet """
        if self.book.is_xlwings:
            self.sheet_obj.api.Unprotect(password)
        else:
            self.sheet_obj.protection.password = password
            self.sheet_obj.protection.disable()
            self.sheet_obj.protection.sheet = False

    def move(self, offset):
        """
         Move the sheet

        :param offset: distance to move
        :return: self
        """
        self.book.move_sheet(self, offset)

    def delete(self):
        """ delete the worksheet """
        self.book.delete_sheet(self)

    def activate(self):
        """ activate the worksheet """
        self.book.active = self

    def __repr__(self):
        return '<WorkSheet %s>' % repr(self.name)


class WorkSheets:
    def __init__(self, book):
        self.book = book

    def __contains__(self, sheet_name):
        return self.book.__contains__(sheet_name)

    def __getitem__(self, item) -> WorkSheet:
        return self.book[item]

    @property
    def count(self):
        return self.book.count

    @property
    def active(self) -> WorkSheet:
        return self.book.active

    @active.setter
    def active(self, value):
        self.book.active = value

    def add(self, name=None, index=None):
        """ add a sheet"""
        self.book.create_sheet(name, index)

    def delete(self, name=None):
        """ delete a worksheet """
        self.book.delete_sheet(name)


# Excel .xlsx 文档类
class Excel(BaseFile):
    def __init__(self, filename=None, template=None, **kwargs):
        """
        init and open file

        :param filename:  the filename to open.
                        if filename is None, create new empty file.
                        if filename is None and run in xlwings, get the workbook that calls,
                        using xlwings.Book.call()

        :param template: (optinal) template filename. If not None, use the template
                        file to create new file.
        :param kwargs:
        """
        self.is_xlwings = False
        self.xlwings_book = None
        self.msg = ''
        self._file_password = None

        # noinspection PyBroadException
        try:
            self.xlwings_book = xlwings.Book.caller()
        except Exception:
            # Exception: Book.caller() must not be called directly. Call through Excel
            pass

        if not self.xlwings_book and filename is None:
            caller_file = inspect.stack()[1].filename
            same_name_xlsx = Util.change_file_ext(caller_file, '.xlsx')
            if os.path.exists(same_name_xlsx):
                filename = same_name_xlsx

        super().__init__(filename, template, **kwargs)

    @property
    def book_obj(self):
        """ Workbook object of openpyxl or book object of xlwings """
        return self._obj

    def _new_file(self, filename, **kwargs):
        if self.xlwings_book:
            self._obj = self.xlwings_book
            self.filename = self.xlwings_book.fullname
            self.is_xlwings = True
            self.msg = '_new_file ' + str(filename)
        else:
            self._obj = openpyxl.Workbook()
            self.is_xlwings = False
            self.filename = None
            # modify 'Sheet' title, make it looks like Microsoft Excel new file
            if 'Sheet' in self._obj:
                self._obj['Sheet'].title = 'Sheet1'
            if filename:
                self.save(filename, **kwargs)
            return self

    def _open(self, filename, **kwargs):
        read_only = kwargs.get('read_only', None)
        data_only = kwargs.get('data_only', None)
        if self.xlwings_book is not None:
            self._obj = self.xlwings_book
            self.is_xlwings = True
        else:
            self._obj = openpyxl.load_workbook(filename, read_only=read_only, data_only=data_only)
            self.is_xlwings = False
            self.filename = filename
        return True

    def _close(self, **kwargs):
        if self._obj is not None:
            save = kwargs.get('save', False)

            # save before close
            if save and self.filename != '':
                self.save()

            # noinspection PyBroadException
            try:
                self._obj.close()
            finally:
                self._obj = None
                self.filename = ''
        return True

    def _save_file(self, filename, **kwargs):
        formats = {
            '.rtf': XlsConsts.wdFormatRTF,
        }

        file_ext = Util.file_ext(filename)

        if file_ext == '.xlsx':
            return self._save_xlsx(filename, **kwargs)

        elif file_ext in formats:
            app, workbook = Win32.open_excel(self.filename, False)
            if not Util.is_absolute_path(filename):
                filename = os.path.join(os.getcwd(), filename)
            workbook.SaveAs(filename, formats[file_ext], **kwargs)
            Win32.close(app, workbook)

        else:
            return super()._save_file(filename, **kwargs)

    def _save_xlsx(self, filename, **kwargs):
        if self._obj is not None:
            filename = self.filename if filename is None or filename == '' else filename
            self._obj.save(filename, **kwargs)
            self.filename = filename

    def __contains__(self, name):
        """ whether worksheet of specified name exists """
        if isinstance(name, WorkSheet):
            name = name.title

        if self.is_xlwings:
            return name in self.sheetnames
        else:
            return name in self.book_obj

    @property
    def sheets(self):
        """ return a collection of worksheets """
        return WorkSheets(self)

    @property
    def sheetnames(self):
        """ return a list of name of worksheets """
        if self.is_xlwings:
            return [sheet.name for sheet in self.book_obj.sheets]
        else:
            return self.book_obj.sheetnames

    @property
    def count(self):
        """ return count of worksheet """
        if self.is_xlwings:
            return self.book_obj.sheets.count
        else:
            return len(self.book_obj.sheetnames)

    def _get_sheet_by_name(self, name: str) -> WorkSheet:
        """ get sheet by name """
        if name in self:
            if self.is_xlwings:
                return WorkSheet(self, self.book_obj.sheets[name])
            else:
                return WorkSheet(self, self.book_obj[name])

    def _get_sheet_by_index(self, index: int) -> WorkSheet:
        """ get sheet object by index"""
        if 0 <= index < self.count:
            if self.is_xlwings:
                return WorkSheet(self, self.book_obj.sheets[index])
            else:
                return WorkSheet(self, self.book_obj.worksheets[index])

    def _get_sheet(self, item, create=True) -> WorkSheet:
        """
        Get worksheet object

        :Chinese: 取得 WorkSheet 对象。

        :param item: （(Optional) tile of sheet, or index of sheet
        :return: return WorkSheet object
        """
        if isinstance(item, WorkSheet):
            return item

        sheet = None

        if isinstance(item, str):
            if item not in self and create:
                sheet = self.create_sheet(item)
            else:
                sheet = self._get_sheet_by_name(item)

        elif isinstance(item, int):
            if item >= self.count and create:
                for n in range(self.count, item + 1):
                    name = 'Sheet' + str(n)
                    count = 1
                    while name in self:
                        name = name + '(' + str(count) + ')'
                        count += 1
                    self.create_sheet(name)
            # find sheet
            sheet = self._get_sheet_by_index(item)

        # return
        if sheet:
            return sheet
        else:
            raise ValueError('WorkSheet %s not exists' % repr(item))

    def create_sheet(self, name=None, index=None):
        """
        Create a new WorkSheet

        :param name:  name of new Sheet
        :param index:  (optional int, index of the worksheet.
            Default if None.(the worksheet will be put at then end)
        :return: self
        """
        if self.is_xlwings:
            sheet_count = self.book_obj.sheets.count
            before = after = None
            if sheet_count > 0:
                if index is None or 0 > index >= sheet_count:
                    after = self.book_obj.sheets[sheet_count - 1]
                elif index == 0:
                    before = self.book_obj.sheets[0]
                else:
                    before = self.book_obj.sheets[index]
            ws = self.book_obj.sheets.add(name, before=before, after=after)
            return WorkSheet(self, ws)
        else:
            ws = self.book_obj.create_sheet(title=name, index=index)
            return WorkSheet(self, ws)

    def move_sheet(self, sheet, offset=0):
        """
        Move the sheet

        :Chinese: 移动WorkSheet

        :param sheet:  worksheet, could be name or WorkSheet object
        :param offset: distance to move
        :return: self
        """
        if self.count == 0:
            return

        sheet = self._get_sheet(sheet)
        count = self.count

        if self.is_xlwings:
            # calculate new index
            index = sheet.index + offset
            if index < 0:
                index = 0

            # move
            if index >= count - 1:
                after = self.sheets[count-1].sheet_obj.api
                sheet.sheet_obj.api.Move(None, after)
            elif 0 <= index < count:
                before = self.sheets[index].sheet_obj.api
                sheet.sheet_obj.api.Move(before)
        else:
            self.book_obj.move_sheet(sheet.sheet_obj, offset=offset)

        return self

    def copy_sheet(self, from_sheet) -> WorkSheet:
        """
        Copy sheet , and create a new worksheet

        :Chinese: 复制 WorkSheet, 创建一个新的WorkSheet

        :param from_sheet: the Worksheet to be copied
        :return: return the new-created Worksheet object
        """
        if isinstance(from_sheet, str):
            from_sheet = self._get_sheet(from_sheet)

        if isinstance(from_sheet, WorkSheet):
            from_sheet = from_sheet.sheet_obj

        ws = self.book_obj.copy_worksheet(from_sheet)
        return WorkSheet(self, ws)

    def delete_sheet(self, sheet):
        """
        Delete a worksheet

        :param sheet: the worksheet name or object
        :return: self
        """
        sheet = self._get_sheet(sheet, False)
        if self.is_xlwings:
            sheet.sheet_obj.delete()
        else:
            self.book_obj.remove(sheet.sheet_obj)

    def index(self, sheet):
        """ retunn the index (int) of the worksheet """
        # noinspection PyBroadException
        try:
            sheet = self._get_sheet(sheet)
            return self.book_obj.index(sheet.sheet_obj)
        except Exception:
            return -1

    @property
    def active(self) -> WorkSheet:
        """ return current activated worksheet """
        if self.is_xlwings:
            return WorkSheet(self, self.book_obj.sheets.active)
        else:
            return WorkSheet(self, self.book_obj.active)

    @active.setter
    def active(self, sheet):
        """ set current activated worksheet """
        sheet = self._get_sheet(sheet)
        if self.is_xlwings:
            sheet.sheet_obj.activate()
        else:
            self.book_obj.active = sheet.sheet_obj

    def protect(self, password: str):
        """
        protect the file

        https://openpyxl.readthedocs.io/en/stable/protection.html

        :param password:  the password
        :return: self
        """
        if self.is_xlwings:
            # https://docs.microsoft.com/en-us/office/vba/api/excel.workbook.protect
            self.book_obj.api.Protect(password, True);
            self.active['A12'].value = password
        else:
            self._obj.security.workbookPassword = password
            self._obj.security.lockStructure = True
        return self

    def unprotect(self, password: str):
        """
        unprotect the file

        :param password: the password
        :return: self
        """
        if self.is_xlwings:
            # https://docs.microsoft.com/en-us/office/vba/api/excel.workbook.unprotect
            self.book_obj.api.Unprotect(password);
        else:
            self._obj.security.workbookPassword = password
            self._obj.security.lockStructure = False
        return self

    def set_file_password(self, password: str, hashed=False):
        if self.is_xlwings:
            self._file_password = password
        else:
            self._obj.security.set_workbook_password(password, already_hashed=hashed)
            return True
        return False

    # noinspection PyMethodMayBeStatic
    def _parse_item(self, item):
        """
        parese the item of __getitem___

        :param item:
        :return: sheet_name, address, row, col, end_row, end_col
        """
        sheet_name = None
        address = None
        row = None
        col = None
        end_row = None
        end_col = None

        if isinstance(item, tuple):
            for n in item:
                if isinstance(n, str):
                    if sheet_name is None:
                        sheet_name = n
                    elif address is None:
                        address = n
                if isinstance(n, int):
                    if row is None:
                        row = n
                    elif col is None:
                        col = n
                    elif end_row is None:
                        end_row = n
                    elif end_col is None:
                        end_col = n
        elif isinstance(item, str):
            offset = item.find('!')
            if offset >= 0:
                sheet_name = item[:offset]
                address = item[offset+1:]
            else:
                sheet_name = item

        return sheet_name, address, row, col, end_row, end_col

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._get_sheet_by_index(item)

        sheet_name, address, row, col, end_row, end_col = self._parse_item(item)

        if sheet_name is None:
            raise ValueError("sheet name should specified")

        if sheet_name not in self:
            self.create_sheet(sheet_name)

        if sheet_name in self:
            sheet = self._get_sheet(sheet_name)
            if address is not None:
                return sheet[address]
            elif row is None and col is None:
                return sheet
            elif end_row is None and end_col is None:
                return sheet.cell(row, col)
            else:
                end_row = row if end_row is None else end_row
                end_col = col if end_col is None else end_col
                addr = [row, col, end_row, end_col]
                return Range(addr, sheet)

    def __repr__(self):
        return '<Excel WorkBook %s>' % repr(self.filename)

    def print(self, **kwargs):
        """
        Call Excel Application,  print the file to printer.

        :Chinese: 调用Excel, 使用打印机打印文档

        :param kwargs:
        :return: self
        """
        print("printing", self.filename, '...')

        if self.is_xlwings:
            self.book_obj.api.PrintOut(**kwargs)
        else:
            app, workbook = Win32.open_excel(self._absolute_filename, False)
            workbook.PrintOut(**kwargs)
            time.sleep(2)
            Win32.close(app, workbook)
        return self

    def find(self, sheet_name, value1, value2=None, regex=False):
        """
        find a cell in specified sheet

        :param sheet_name: worksheet name
        :param value1:   first value
        :param value2:  (optional) second value
        :param regex:   (optional) whether use regex

        :return: return Cell object if found. return None if not found.
        """
        if sheet_name in self:
            sheet = self._get_sheet(sheet_name)
            return sheet.find(value1, value2, regex)



import office


def example_open_read_write_cells():
    # 打开 datafile.xlsx 文件，返回 workbook 对象
    workbook = office.open_file("datafile.xlsx")

    # 如果 workbook 中有名为 '商品销售' 的工作表
    if '商品销售' in workbook:
        sheet = workbook['商品销售']  # 取得工作表对象
        cell = sheet['B2']  # 取得 B2 单元格对象
        print(cell.value)  # 打印 B1 单元格的值
        cell.value = '新城'  # 修改 B2 单元格的值

        range1 = sheet['B2:C4']  # 取得 B2:C4 区域
        print(range1.value)  # 打印 区域的值

    workbook.save()
    workbook.close()


def example_direct_read_write():
    # 打开 datafile.xlsx 文件，返回 workbook 对象
    wb = office.open_file("datafile.xlsx")

    # 通过workbook对象直接获取 商品销售!B2 单元格对象
    # 如果 商品销售 工作表不存在，则自动创建工作表
    print(wb['商品销售!B2'].value)

    # 通过workbook对象直接写入 商品销售!B2 单元格的值
    # 如果 商品销售 工作表不存在，则自动创建工作表
    wb['商品销售2!B2'].value = "上海"

    print(wb['商品销售!B2'].value)  # 显示该单元格已写入 "上海"

    # 通过workbook对象直接获取 商品销售!B2:C4 区域对象
    range1 = wb['商品销售!B2:C4']
    # 区域对象的 values 属性返回一个列表， 包含区域内所有单元格的值
    print(range1.values)

    # 通过workbook对象获取 商品销售!B2:C4 区域对象, values 写入各单元格的值
    wb['商品销售!B21:C22'].values = [['北京', 18], ['上海', 20]]

    print(wb['商品销售!B21:C22'].values)  # 显示该区域已写入

    wb.save()
    wb.close()


def example_find():
    wb = office.open_file("datafile.xlsx")

    # 查找 商品销售 工作表中 单元格值为 "中山" 的单元格
    # cell = wb.find("商品销售", "中山")
    # print(cell)

    # 查找商品销售表中 "中山" 与 "金额"的交叉单元格，即：中山的金额
    cell = wb.find("商品销售", "中山", "金额")
    if cell:
        print(cell.value)


def find_file(month, sheet_name, area, item):
    """ 查找文件 """
    wb = office.open_file(str(month) + '.xlsx')
    cell = wb.find(sheet_name, area, item)
    if cell:
        return cell.value


def example_expand():
    wb = office.open_file("datafile.xlsx")

    # 取得一个单元格  商品销售!A1
    cell = wb["商品销售"]['A1']

    # 扩展单元格, 返回一个区域, 返回结果是 A1:D6
    print(cell.expand())

    # 扩展单元格(水平方向不扩展), 返回一个区域, 返回结果是 A1:A6
    print(cell.expand(right=None))

    # 扩展单元格(垂直方向不扩展), 返回一个区域, 返回结果是 A1:D1
    print(cell.expand(down=None))


def example_range_to_dataframe():
    wb = office.open_file("datafile.xlsx")

    # 取得一个区域  商品销售!A1:D6
    range1 = wb["商品销售", 'A1:D6']
    print(range1)

    # 取得该区域的数据，返回list
    list1 = range1.to_list()
    print(list1)

    # 取得该区域的数据，返回 DataFrame
    df = range1.to_dataframe()
    print(df)

    # 写入DataFrame数据
    range1.values = df

    # 写入list数据
    range1.values = list1

    # wb.save()
    # wb.close()


if __name__ == "__main__":
    # example_open_read_write_cells()
    # example_direct_read_write()
    # example_find()
    # print(find_file(202201, '商品销售', "中山", "金额"))
    # example_expand()
    example_range_to_dataframe()



import math
from .util import Util

try:
    import PIL
    from PIL import Image, ImageDraw, ImageEnhance, ImageFont
except ImportError:
    PIL = None
    Image = None
    ImageDraw = None
    ImageEnhance = None
    ImageFont = None


try:
    import numpy
except ImportError:
    numpy = None


class PhotoFont:
    def __init__(self, name='arial', size=20, color='#000'):
        self._name = name
        self._font_file = name + '.ttf'
        self._size = size
        self._color = color
        self._create_obj()

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value
        self._font_file = self._name + '.ttf'
        self._create_obj()

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, value):
        self._size = value
        self._create_obj()

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, value):
        self._color = value

    def _create_obj(self):
        self.font_obj = ImageFont.truetype(self._font_file, self._size)

    # noinspection PyMethodMayBeStatic
    def _get_font_name(self, text: str):
        """ 取得字体名称 """
        if Util.has_chinese(text):
            return 'simhei.ttf'
        else:
            return 'arial.ttf'

    def set_text(self, text):
        self._font_file = self._get_font_name(text)
        self._create_obj()


class PhotoSplitter:
    def __init__(self, photo):
        Util.check_module(PIL, "pillow")
        self.photo = photo
        self.boxes = []
        self.index = 0

    def table(self, rows, cols, spacing=0, border=0):
        """
        划分表格

        :param rows:  表格的行数 (Pixels)
        :param cols:  表格的列数 (Pixels)
        :param spacing: 间隔宽度 (Pixels)
        :param border: 边框宽度 (Pixels)
        :return: self
        """
        width, height = self.photo.size
        spacing = spacing if spacing > 0 else 0

        if not isinstance(rows, int) or not isinstance(cols, int):
            raise ValueError("rows cols should be int")

        # 每个单元表的宽、高
        w = round((width - (cols-1) * spacing - 2 * border) / cols)
        h = round((height - (rows-1) * spacing - 2 * border) / rows)
        if w < 1 or h < 1:
            raise ValueError("cannot split into rows %s and cols %s" % (rows, cols))

        # 清空原数据
        self.boxes = []
        self.index = 0

        # 切分表格
        x = y = border
        for row in range(0, rows):
            for col in range(0, cols):
                rect = (x, y, x + w, y + h)
                self.boxes.append(rect)
                x = x + spacing + w
            y = y + spacing + h
            x = border

        return self

    def table_fill(self, photo, index=None):
        """ fill a photo into current cell """
        if len(self.boxes) == 0:
            raise ValueError("please call table() first")

        index = self.index if index is None else index

        if index >= len(self.boxes) or index < 0:
            index = 0

        if index < len(self.boxes):
            box = self.boxes[index]
            self.photo.draw_photo(photo, box)
            self.index = index + 1 if index < len(self.boxes) else 0
        else:
            raise IndexError("index %s out of bound" % index)


class Photo:
    """

    Photo is an extended class of PIL.Image.


    PIL Image document:
    https://pillow.readthedocs.io/en/stable/reference/Image.html#module-PIL.Image

    """

    # modes
    BIT = '1'
    PALETTE = 'P'
    GRAYSCALE = 'L'
    RGB = 'RGB'
    RGBA = 'RGBA'
    CYMK = 'CYMK'
    YCbCr = 'YCbCr'
    LAB = 'LAB'
    HSV = 'HSV'

    # resize styles
    FIT = "fit"
    STRETCH = "stretch"
    GROW = "grow"
    TILE = "tile"

    # maximum width, height of the photo
    MAX_WIDTH = 50000

    def __init__(self, file_or_size, height=None, **kwargs):
        """
        init instance

        :param file_or_size:  A filename (string), or file object, or size tuple
        :param kwargs: Extra parameters, such as 'mode', 'color'
        """
        self.filename = None
        self._font = PhotoFont()
        self._draw_obj = None
        self._splitter = None
        self.base_obj = None

        color = kwargs.get('color', (255, 255, 255))
        mode = kwargs.get('mode', 'RGB')

        if isinstance(file_or_size, str):
            base_obj = Image.open(file_or_size)
            self.filename = file_or_size
        elif isinstance(file_or_size, tuple):
            base_obj = Image.new(mode, file_or_size, color)
        elif isinstance(file_or_size, int):
            width = file_or_size
            if not isinstance(height, int):
                height = width
            base_obj = Image.new(mode, (width, height), color)
        elif isinstance(file_or_size, Image):
            base_obj = file_or_size
        else:
            raise ValueError("unsupport file type %s" % repr(type(file_or_size)))

        super().__init__(base_obj)

        # decorate method of Images
        # self.resize = self._decorate_method(self.base_obj.resize, Image)

    @property
    def im(self):
        return self.base_obj

    def _decorate_method(self, method, cls=None):
        that = self

        def warpper(*args, **kwargs):
            params = args[1:]
            result = method(*params, **kwargs)
            if cls and isinstance(result, cls):
                that.base_obj = result
            return that

        return warpper

    @property
    def width(self):
        return self.base_obj.width

    @property
    def height(self):
        return self.base_obj.height

    @property
    def mode(self):
        return self.base_obj.mode

    @property
    def size(self):
        return self.base_obj.size

    def __repr__(self):
        s = "<Photo %s x %s, %s>" % (self.width, self.height, repr(self.mode))
        return s

    def new(self, width, height, color=(255, 255, 255), mode='RGB'):
        """
        Creates a new image with the given mode and size.
        https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.new

        :param width: widtd in pixels.
        :param height: height in pixels.
        :param mode:  The mode to use for the new image, such as:  Photo.RGB, Photo.RGBA, ...
        :param color: What color to use for the image. Default is black. If given, this should be a single integer
        or floating point value for single-band modes, and a tuple for multi-band modes (one value per band).
        :return: self
        """
        self.base_obj = Image.new(mode, (width, height), color)
        return self

    def open(self, file, formats=None):
        """
        Opens and identifies the given image file.
        https://pillow.readthedocs.io/en/stable/reference/Image.html#functions

        :param file:  A filename (string), pathlib.Path object or a file object.
        :param formats: A list or tuple of formats to attempt to load the file in.
        :return: self
        """
        self.base_obj = Image.open(file, 'r', formats)
        return self

    def save(self, file=None, fmt=None, **params):
        """
        Saves this image under the given filename. If no format is specified, the format to use is determined
        from the filename extension, if possible.
        https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.save

        :param file:   A filename (string), pathlib.Path object or file object.
        :param fmt: Optional format override. If omitted, the format is determined from the filename extension.
        :param params: Extra parameters to the image writer.
        :return: self
        """
        file = self.filename if file is None else file
        try:
            self.base_obj.save(file, fmt, **params)
            if isinstance(file, str):
                self.filename = file
        except Exception as e:
            err_msg = str(e)
            if err_msg.find('RGBA'):
                self.base_obj = self.base_obj.convert('RGB')
                self.base_obj.save(file, fmt, **params)
                if isinstance(file, str):
                    self.filename = file
        return self

    def clone(self) -> 'Photo':
        """ return a new copy of the photo"""
        new_photo = Photo(self.width, self.height, mode=self.mode)
        new_photo.paste(self)
        return new_photo

    def table_split(self, rows, cols, spacing=0, border=0):
        """ 切分为 多张小图，返回图片列表 """
        splitter = PhotoSplitter(self).table(rows, cols, spacing, border)
        result = []
        for box in splitter.boxes:
            p = self.clone().crop(box)
            result.append(p)
        return result

    def table(self, rows, cols, spacing=0, border=0):
        """
        切分为表格

        :param rows:  表格的行数 (Pixels)
        :param cols:  表格的列数 (Pixels)
        :param spacing: 间隔宽度 (Pixels)
        :param border: 边框宽度 (Pixels)
        :return: self
        """
        self._splitter = PhotoSplitter(self).table(rows, cols, spacing, border)
        return self

    def table_view(self, rows, cols, spacing=0, border=0, fill="#FFF"):
        """
        将图片变成表格状

        :param rows:  表格的行数 (Pixels)
        :param cols:  表格的列数 (Pixels)
        :param spacing: (可选)间隔宽度 (Pixels)
        :param border: (可选)边框宽度 (Pixels)
        :param fill:   (可选)底色
        :return:
        """
        im = self.base_obj
        photos = self.table_split(rows, cols, spacing, border)
        p = Photo(im.size, mode=im.mode, color=fill)
        p.table(rows, cols, spacing, border).table_fill(photos)
        self.base_obj = p.base_obj
        return self

    def table_fill(self, photo, index=None):
        """
        table() 切分为表格后, 填入图片

        :param photo: 图片， Photo对象 或 图片文件名
        :param index: （可选) 表格位置
        :return: self
        """

        if not self._splitter:
            raise ValueError("please call table() first")

        if isinstance(photo, list):
            self._splitter.index = 0
            for p in photo:
                self._splitter.table_fill(p)
        else:
            self._splitter.table_fill(photo, index)

        return self

    def border(self, color, left, top=None, right=None, bottom=None):
        """
        扩大图片，创建边框

        :param color:  边框颜色
        :param left:   左边边框宽度
        :param top:    （可选)上边边框宽度
        :param right:  （可选)右边边框宽度
        :param bottom: （可选)下边边框宽度
        :return: self
        """
        right = left if right is None else right
        top = left if top is None else top
        bottom = top if bottom is None else bottom
        im = self.base_obj
        w, h = im.size
        p = Photo(w + left + right, h + top + bottom, mode=im.mode, color=color)
        box = (left, top, left + w, top + h)
        p.paste(self, box)
        self.base_obj = p.base_obj
        return self

    def _validate_box(self, box, default_width=None, default_height=None):
        """ validate the box """
        if box is None:
            box = (0, 0, self.width, self.height)

        if default_width is None:
            default_width = self.width

        if default_height is None:
            default_height = self.height

        if isinstance(box, tuple):
            if len(box) == 2:
                right = min(box[0]+default_width, self.width)
                bottom = min(box[1]+default_height, self.height)
                box = tuple(list(box).append(right, bottom))
            if len(box) == 4:
                return box

        raise ValueError('box should be a 4-int tuple')

    # noinspection PyMethodMayBeStatic
    def _validate_wh(self, width, height, msg=None):
        """ validate width and height """
        if not isinstance(width, int) or not isinstance(height, int):
            raise ValueError('width and height should be int')

        if width <= 0 or height <= 0:
            msg = 'width or height should large than zero' if msg is None else msg
            raise ValueError(msg)

        if width > Photo.MAX_WIDTH or height > Photo.MAX_WIDTH:
            raise ValueError('width or height is too big')

    # noinspection PyMethodMayBeStatic
    def _get_image(self, photo):
        """ get image obj from photo"""
        if isinstance(photo, Photo):
            return photo.base_obj
        else:
            return photo

    # noinspection PyMethodMayBeStatic
    def _get_photo(self, im):
        """ get image obj from photo"""
        return im if isinstance(im, Photo) else Photo(im)

    def histogram(self, channel='L', mask=None, extrema=None):
        """
        取得直方图数据

        :param channel:  通道， R , G, B, L(灰度), 默认是 L
        :param mask:     （可选)掩膜
        :param extrema:   （可选)
        :return: 返回一个数据列表

        https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.histogram

        """
        mode = str(self.mode).upper()
        channel = str(channel).upper()

        if channel == 'L':
            # convert to grayscale
            p = self.base_obj.convert(Photo.GRAYSCALE)
            return p.histogram(mask, extrema)
        elif mode.find(channel) >= 0:
            offset = mode.find(channel)
            # split and find channel
            channels = self.base_obj.split()
            if offset < len(channels):
                return channels[offset].histogram(mask, extrema)
        elif channel == '' or channel is None:
            # all channel mixed
            return self.im.histogram(mask, extrema)
        else:
            raise ValueError('channel %s not found' % repr(channel))

    def crop(self, box):
        """
        cut down a rectangular region from this image

        SEE: https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.crop

        :param box: the crop rectangle, as a (left, upper, right, bottom)-tuple.
        :return: self
        """
        if isinstance(box, tuple) and len(box) == 4:
            self.base_obj = self.base_obj.crop(box)
        else:
            raise ValueError("box should be a 4-tuple of (left, top, right, bottom) ")
        return self

    def flip(self):
        """ 图片左右反转"""
        self.base_obj = self.base_obj.transpose(PIL.Image.Transpose.FLIP_LEFT_RIGHT)
        return self

    def vflip(self):
        """ 图片上下反转"""
        self.base_obj = self.base_obj.transpose(PIL.Image.Transpose.FLIP_TOP_BOTTOM)
        return self

    def rotate(self, degree, fill=None, expand=0, center=None, resample=None, translate=None):
        """
        旋转图像

        https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.rotate

        :param degree:   角度， 0 - 360度
        :param fill:     旋转后产生的空白部分的填充颜色, 默认是黑色
        :param expand:   是否扩大图片，以保证原图全部可见
        :param center:   旋转中心点， tuple (x, y)
        :param resample: 重采样方法
        :param translate: An optional post-rotate translation
        :return:
        """
        resample = PIL.Image.Resampling.BICUBIC if resample is None else resample

        if degree in [-90, -180, -270, -360]:
            degree += 360

        if degree == 0:
            pass
        elif degree == 90:
            self.base_obj = self.base_obj.transpose(PIL.Image.Transpose.ROTATE_90)
        elif degree == 180:
            self.base_obj = self.base_obj.transpose(PIL.Image.Transpose.ROTATE_180)
        elif degree == 270:
            self.base_obj = self.base_obj.transpose(PIL.Image.Transpose.ROTATE_270)
        else:
            self.base_obj = self.base_obj.rotate(degree, resample, expand, center, translate, fill)
        return self

    def resize(self, width, height=None, style='fit', **kwargs):
        """
        resized the image

        :param width:   An int value which is the target width of the photo ,
                        or a float value which is scale ratio
        :param height:  (optional)An int which is target height of the photo.
                        or None when width is float
        :param style:  (optional) A str, style could be Photo.STRETCH, Photo.FIT, Photo.GROW,
                       Photo.STRETCH  means stretch the photo to match the target size.
                       Photo.FIT  means cut down the edges of the photo to match the target size.
                       Photo.GROW  means add edges to the photo to match the target size.
        :param kwargs: (optional) accept resample, reducing_gap, box
                SEE: https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.resize
        :return: self
        """
        # get size`
        old_w, old_h = self.size
        self._validate_wh(old_w, old_h, 'photo is empty')
        old_ratio = old_w / old_h

        # get params
        resample = kwargs.pop("resample", None)
        reducing_gap = kwargs.pop("reducing_gap", None)
        scale = kwargs.pop("scale", False)
        box = kwargs.pop("box", None)

        # process the width, height
        if isinstance(width, float) and height is None:
            # width param is a scale ratio
            new_ratio = width
            width = round(self.width * new_ratio)
            height = round(self.height * new_ratio)
            self._validate_wh(width, height)
            self.base_obj = self.base_obj.resize((width, height), resample, reducing_gap)
            return self

        # if width, height is single provided, keep ratio
        if isinstance(width, int) and height is None:
            height = round(width / old_ratio)
        elif isinstance(height, int) and width is None:
            width = round(height * old_ratio)

        self._validate_wh(width, height)
        new_ratio = width / height

        style = str(style).lower().strip()
        if style not in [Photo.STRETCH, Photo.FIT, Photo.GROW]:
            raise ValueError("unknown style %s" % repr(str(style)))

        # if box is specified, crop it
        if box is not None:
            self.crop(box)

        if style == Photo.STRETCH:
            # stretch the photo to match the target size.
            self.base_obj = self.base_obj.resize((width, height), resample, reducing_gap)
        elif style == Photo.FIT:
            # FIT style means cut down the edges of photo outside the target size
            x = y = 0
            if new_ratio < old_ratio:
                new_h = old_h
                new_w = round(new_h * new_ratio)
                x = round((old_w - new_w)/2)
            else:
                new_w = old_w
                new_h = round(new_w / new_ratio)
                y = round((old_h - new_h) / 2)
            box = (x, y, x + new_w, y + new_h)
            self.base_obj = self.base_obj.crop(box)
            if not scale:
                self.base_obj = self.base_obj.resize((width, height), resample, reducing_gap)
        elif style == Photo.GROW:
            # GROW style is to keep all photo, add margins if needed
            if new_ratio > old_ratio:
                new_h = old_h
                new_w = round(new_h * new_ratio)
            else:
                new_w = old_w
                new_h = round(new_w / new_ratio)
            if not scale:
                self.base_obj = self.base_obj.resize((width, height), resample, reducing_gap)
            # create new photo
            new_photo = Photo((new_w, new_h), img=self.base_obj.mode)
            old_photo = self
            x = round((new_photo.width - old_photo.width)/2)
            y = round((new_photo.height - old_photo.height) / 2)
            # paste old_photo into the new photo
            new_photo.paste(old_photo, (x, y))
            self.base_obj = new_photo.base_obj
        return self

    def thumbnail(self, width, height=None, style=None, resample=None, reducing_gap=2.0):
        """
        生成缩略图

        https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.thumbnail

        :param width:  缩略图宽度
        :param height: 缩略图高度
        :param style:  (optional) A str, style could be Photo.STRETCH, Photo.FIT, Photo.GROW,
                       Photo.STRETCH  means stretch the photo to match the target size.
                       Photo.FIT  means cut down the edges of the photo to match the target size.
                       Photo.GROW  means add edges to the photo to match the target size.
        :param resample: 重采样方法
        :param reducing_gap:
        :return:
        """
        resample = PIL.Image.Resampling.BICUBIC if resample is None else resample

        if height is None and isinstance(width, int):
            height = round(width * (self.height / self.width))
        elif width is None and isinstance(height, int):
            width = round(height * (self.width / self.height))

        if style:
            kwargs = {}
            kwargs['resample'] = resample
            kwargs['reducing_gap'] = reducing_gap
            self.resize(width, height, style, **kwargs)
        else:
            self.base_obj.thumbnail((width, height), resample, reducing_gap)

        return self

    def alpha(self, value):
        """
        Adds or replaces the alpha layer in this image. If the image does not have an alpha layer, it’s converted
        to “LA” or “RGBA”. The new layer must be either “L” or “1”.

        :param value: alpha – The new alpha layer. This can either be an “L” or “1” image having the same size
        as this image, or an integer or other color value.

        :return: self
        """
        self.base_obj.putalpha(value)
        return self

    def show(self, title=None):
        """
        Displays this image.

        https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.show

        :param title: window title
        :return: self
        """
        self.base_obj.show(title)
        return self

    # noinspection PyMethodMayBeStatic
    def _angle(self, width, height):
        """ return angle in degree """
        if height == 0:
            return 0
        else:
            return math.atan(height / width) * 180 / math.pi

    def blend(self, photo, alpha):
        """
        Creates a new image by interpolating input image, using a constant alpha.

        https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.blend

        :param photo: The image. Must have the same mode and size as the first image.
        :param alpha: The interpolation alpha factor. If alpha is 0.0, a copy of the first image is returned.
        If alpha is 1.0, a copy of the second image is returned. There are no restrictions on the alpha value.
        If necessary, the result is clipped to fit into the allowed output range.

        :return: self
        """
        self.base_obj = PIL.Image.blend(self.base_obj, self._get_image(photo), alpha)
        return self

    def composite(self, photo, mask):
        """
        Create composite image by blending images using a transparency mask.

        https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.composite

        :param photo:  The image. Must have the same mode and size as the first image.
        :param mask: A mask image. This image can have mode “1”, “L”, or “RGBA”, and must have
        the same size as the other two images.

        :return: self
        """
        self.base_obj = PIL.Image.composite(self.base_obj, self._get_image(photo), mask)
        return self

    def eval(self, func):
        """
        Applies the function (which should take integer argument) to each pixel in the given image.

        :param func:  the function (which should take integer argument)
        :return: self
        """
        self.base_obj = PIL.Image.eval(self.base_obj, func)
        return self

    def convert(self, mode):
        """
        convert to specified mode

        :param mode: the mode to use for the new image, such as:  Photo.RGB, Photo.RGBA, ...
        :return: self
        """
        self.base_obj = self.base_obj.convert(mode)
        return self

    def transparent(self, rgb_color):
        """
        设置透明色

        :param rgb_color:  颜色， tuple like (R, G, B)
        :return: self
        """
        if self.mode != 'RGBA':
            self.convert('RGBA')

        Util.check_module(numpy, 'numpy')

        if numpy:
            # Make transparency using numpy
            arr = numpy.array(self.base_obj)  # convert image to numpy array
            rgb = rgb_color
            mask = (arr[:, :, 0] == rgb[0]) & (arr[:, :, 1] == rgb[1]) & (arr[:, :, 2] == rgb[2])
            arr[mask] = [255, 255, 255, 0]
            # convert the numpy array into image
            self.base_obj = PIL.Image.fromarray(arr)
        else:
            # make transparency using python
            arr = []
            for item in self.base_obj.getdata():
                if item[:3] == rgb_color:
                    arr.append((255, 255, 255, 0))
                else:
                    arr.append(item)
            self.base_obj.putdata(arr)
        return self

    def paste(self, photo, box=None, mask=None):
        """
        Pastes another image into this image

        :param photo:  Source image or pixel value (integer or tuple).
        :param box:    An optional 4-tuple giving the region to paste into. If a 2-tuple is used instead, it’s
        treated as the upper left corner. If omitted or None, the source is pasted into the upper left corner.

        If an image is given as the second argument and there is no third, the box defaults to (0, 0), and the
        second argument is interpreted as a mask image.
        :param mask:   An optional mask image.
        :return:
        """
        self.base_obj.paste(self._get_image(photo), box, self._get_image(mask))
        return self

    def brightness(self, factor):
        """
        调节亮度

        :param factor : 调节系数 float型. 0调至全黑，1 保持原图亮度不变，>1提高亮度, <1降低亮度
        :return: self
        """
        en = ImageEnhance.Brightness(self.base_obj)
        self.base_obj = en.enhance(factor)
        return self

    def color(self, factor):
        """
        调节色彩浓度

        :param factor : 调节系数 float型. 0调至黑白，1 保持原图色彩浓度不变，>1提高色彩浓度, <1降低色彩浓度
        :return: self
        """
        en = ImageEnhance.Color(self.base_obj)
        self.base_obj = en.enhance(factor)
        return self

    def contrast(self, factor):
        """
        调节对比度

        :param factor : 调节系数 float型. 0调至全灰(无对比度)，1 保持原图不变，>1提高对比度, <1降低对比度
        :return: self
        """
        en = ImageEnhance.Contrast(self.base_obj)
        self.base_obj = en.enhance(factor)
        return self

    def sharpness(self, factor):
        """
        调节锐度

        :param factor : 调节系数 float型. <0是模糊化， >0是锐化
        :return: self
        """
        en = ImageEnhance.Sharpness(self.base_obj)
        self.base_obj = en.enhance(factor)
        return self

    @property
    def font(self):
        return self._font

    @property
    def _draw(self):
        """ return the PIL draw object """
        if self._draw_obj is None or self._draw_obj.im != self.base_obj:
            self._draw_obj = ImageDraw.Draw(self.base_obj)
        return self._draw_obj

    def text_size(self, text):
        """ return a tuple (width, height) of the drawing size of text """
        self.font.set_text(text)
        _, _, width, height = self._draw.textbbox((0, 0), text, self.font.font_obj)
        return width, height

    def draw_photo(self, photo: 'Photo', box=None, style="fit", mask=None):
        """ draw a image """
        photo = self._get_photo(photo)
        box = self._validate_box(box, photo.width, photo.height)
        w = box[2] - box[0]
        h = box[3] - box[1]
        p = photo.clone().resize(w, h, style)
        self.paste(p, box=box, mask=mask)
        return self

    def draw_text(self, x, y, text):
        """ draw text on the photo"""
        self.font.set_text(text)
        self._draw.text((x, y), text, font=self.font.font_obj, fill=self.font.color)
        return self

    def watermark(self, mark, box=None):
        """ draw watermark """
        if isinstance(mark, str):
            width, height = self.base_obj.size
            m = Photo((width, height), color=(255, 255, 255))
            m.font.size = 80
            m.font.size = round(min(width, height) / max(len(mark), 6))
            m.font.color = '#666'
            w, h = m.text_size(mark)
            x = (width - w) / 2
            y = (height - h) / 2
            m.draw_text(x, y, mark)
            m.base_obj = m.base_obj.rotate(30, fillcolor=(255, 255, 255))
            m.save("text.png")
            mark = m

        if isinstance(mark, Photo):
            mark.alpha(100).transparent((255, 255, 255))
            self.paste(mark, box=box, mask=mark)
        return self

    # noinspection PyMethodMayBeStatic
    def lost_red(self, pixels, x, y, channels):
        a = 255
        if channels == 3:
            r, g, b = pixels[x, y]
        else:
            r, g, b, a = pixels[x, y]

        if channels == 3:
            return r, 0, b
        else:
            return r, 0, b, a

    def process_pixels(self, func):
        im = self.base_obj
        width, height = im.size
        pixels = im.load()
        channels = len(pixels[0, 0])
        for x in range(width):
            for y in range(height):
                pixels[x, y] = func(pixels, x, y, channels)



#
# Word : auto creation of the .docx file
#
# Example:
# ::
#     import office
#
#     # create file output.docx using template1.docx
#     doc = office.open_file("output.docx", template="template1.docx")
#
#     # create document content by fill the data from Excel file datafile.xlsx, and save.
#     doc.fill('datafile.xlsx').save()
#
#     # save to pdf with watermark (works on Windows with Microsoft PowerPoint)
#     doc.save('final.pdf', watermark="CONFIDENTIAL")
#
#

import os
import time
import copy
import docx
from types import MethodType
from .base import BaseFile
from .util import Util
from .variable import Variable, VarData
from .win32 import Win32
from .pdf import PDF


class WordConsts:
    # https://docs.microsoft.com/en-us/office/vba/api/word.wdsaveformat
    wdFormatDocument = 0   # Microsoft Office Word 97 - 2003 binary file format.
    wdFormatTemplate = 1  # Word template format.
    wdFormatText = 2  # Microsoft Windows text format.
    wdFormatRTF = 6  # Rich text format (RTF).
    wdFormatHTML = 8  # Standard HTML format.
    wdFormatWebArchive = 9  # Web archive format.
    wdFormatFilteredHTML = 10  # Filtered HTML format.
    wdFormatXMLTemplate = 14  # XML template format.
    wdFormatPDF = 17  # PDF format.


# Word .docx 文档类
class Word(BaseFile):
    """ Word 文档类 """

    class Paragraphs:
        """ 段落集合类 """
        def __init__(self, word):
            self.word = word
            # noinspection PyProtectedMember
            self.document = word._obj

        def __len__(self):
            return len(self.document.paragraphs)

        def _add_method(self, paragraph):
            """ add method to page object """
            def move_to(this, new_index):
                this.word.move_paragraph(this.get_index(), new_index)

            def get_index(this):
                return this.element.getparent().index(this.element)

            def delete(this):
                this.word.delete_paragraph(this.get_index())

            # noinspection PyProtectedMember
            paragraph.element = paragraph._element
            paragraph.word = self.word
            paragraph.move_to = MethodType(move_to, paragraph)
            paragraph.delete = MethodType(delete, paragraph)
            paragraph.get_index = MethodType(get_index, paragraph)
            return paragraph

        def __getitem__(self, item):
            if 0 <= item < len(self.document.paragraphs):
                p = self.document.paragraphs[item]
                return self._add_method(p)
            raise IndexError('Index %s out of bound' % repr(item))

    def _open(self, filename, **kwargs):
        self._obj = docx.Document(filename)
        self.filename = filename
        return True

    def _close(self):
        self._obj = None
        self.filename = None
        return True

    def _new_file(self, filename, **kwargs):
        self._obj = docx.Document()
        if filename:
            self.save(filename)

    def _save_file(self, filename, **kwargs):
        formats = {
            '.txt': WordConsts.wdFormatRTF,
            '.rtf': WordConsts.wdFormatText,
            '.dot': WordConsts.wdFormatTemplate,
            '.dotx': WordConsts.wdFormatXMLTemplate,
            '.doc': WordConsts.wdFormatDocument,
            '.pdf': WordConsts.wdFormatPDF,
            '.html': WordConsts.wdFormatHTML,
            '.htm': WordConsts.wdFormatHTML,
            '.mht': WordConsts.wdFormatWebArchive,
            '.mhtml': WordConsts.wdFormatWebArchive,
        }

        file_ext = Util.file_ext(filename)

        if file_ext == '.docx':
            self._obj.save(filename, **kwargs)
            self.filename = filename
            return True

        elif file_ext in formats:
            app, document = Win32.open_word(self.filename, False)
            if not Util.is_absolute_path(filename):
                filename = os.path.join(os.getcwd(), filename)

            password = kwargs.pop('password', None)
            watermark = kwargs.pop('watermark', None)

            document.SaveAs(filename, formats[file_ext], **kwargs)
            Win32.close(app, document)

            if (password or watermark) and Util.file_ext(filename) == '.pdf':
                PDF(filename).watermark(watermark).save(password=password)

        else:
            return super()._save_file(filename, **kwargs)

    def _fill_data(self, data):
        """
        将数据填入, 替换变量标签，生成文件内容。

        :param data:  dictionary 或 Workbook object 或 DataFrame object
        :return: self
        """
        row = 0
        level = 0

        # 逐个段落处理
        i = 0
        while i < len(self._obj.paragraphs):
            paragraph = self.paragraphs[i]
            i += 1
            result, level = self._fill_paragraph(paragraph, data, row, level, ignore=False)
            if hasattr(result, 'element'):  # 发现 @repeat 时，result 是 Paragraph 对象
                # 处理复制段落
                i = self._repeat_paragraph(i - 1, data)
                level = 0

        return self

    def _repeat_paragraph(self, index, var_data):
        # 1，发现 @repeat.
        # 2, 找到相关段落、变量名列表
        # 3，找到变量的最大数量N
        # 4，复制段落（备份）N 次
        # 5，处理复制段落(忽略@repeat)

        # 找到需要重复的段落、变量名列表
        indexes, var_names = self._find_repeat_paragraphs(index)
        # 找到变量的最大行数量 repeat_count
        repeat_count = 0
        for v in var_names:
            count = var_data.var_length(v)
            count = 1 if not isinstance(count, int) else count
            if count > repeat_count:
                repeat_count = count

        if repeat_count == 0:
            # 如果repeat_count=0，删除段落
            for _ in indexes:
                self.delete_paragraph(index)
            return index
        else:
            # 如果repeat_count>0， 复制段落 repeat_count 次
            copied = 0
            for n in range(repeat_count - 1):
                for idx in indexes:
                    new_index = index + len(indexes) + copied
                    self.copy_paragraph(idx, new_index)
                    copied += 1

            # 处理重复段落(忽略@repeat)
            for row in range(1, repeat_count + 1):
                level = 0
                for line in range(len(indexes)):
                    i = index + (row - 1) * len(indexes) + line
                    paragraph = self.paragraphs[i]
                    result, level = self._fill_paragraph(paragraph, var_data, row, level, ignore=True)
            # 返回重复段落后的段落
            index = index + copied + len(indexes)
        return index

    def _insert_object(self, paragraph, filename):
        pass

    def _fill_paragraph(self, paragraph, data: VarData, row, level, ignore=False):
        """ 将数据填入 paragraph """

        def process(var):
            """ 处理一个变量 """
            value = data.get_var(var, row)
            value = str(value) if value is not None else ''
            if value.startswith(Variable.OBJECT_PREFIX):  # 如果 value 是对象
                # noinspection PyBroadException
                try:
                    # 插入对象
                    self._insert_object(paragraph, value[len(Variable.OBJECT_PREFIX):])
                except Exception:
                    pass
                return ''
            else:
                return value

        varname = ''
        for run in paragraph.runs:
            old_text = run.text
            text = ''
            for idx, c in enumerate(old_text):
                if c == "{":
                    varname += c
                    level += 1
                elif c == "}":
                    varname += c
                    level -= 1
                    if level == 0:
                        if data.has_var(varname):
                            val = process(varname)
                            text += val
                        else:
                            text += varname
                        varname = ''
                else:
                    if level > 0:
                        varname += c
                        if varname == '{@repeat':
                            if ignore:
                                varname = ''
                                level = level - 1
                            else:
                                run.text = text + varname
                                return paragraph, level
                    else:
                        text += c

            run.text = text
        return None, level

    def _find_repeat_paragraphs(self, i):
        """ return (list_paragraph_object, list_varname) """
        varname = ''
        level = 0
        result = []
        var_names = []
        for i in range(i, len(self.paragraphs)):
            paragraph = self.paragraphs[i]
            result.append(i)
            for run in paragraph.runs:
                for idx, c in enumerate(run.text):
                    if c == "{":
                        varname = c
                        level += 1
                    elif c == "}":
                        varname += c
                        level -= 1
                        if Variable.is_var(varname):
                            var_names.append(varname)
                        varname = ''
                        if level == 0:
                            return result, var_names
                    else:
                        if level > 0:
                            varname += c
                        else:
                            pass
        return result, var_names

    @property
    def paragraphs(self):
        return Word.Paragraphs(self)

    @staticmethod
    def _find_paragraph(doc_obj, paragraph):
        for index, p in enumerate(doc_obj.paragraphs):
            # noinspection PyProtectedMember
            if p._element == paragraph._element:
                return index
        return -1

    def find_paragraph(self, paragraph):
        return Word._find_paragraph(self._obj, paragraph)

    def copy_paragraph(self, index, new_index=None, dst_doc=None):
        if dst_doc is None:
            dst_doc = self._obj
        elif isinstance(dst_doc, Word):
            dst_doc = dst_doc._obj

        # create paragraph
        src = self._obj.paragraphs[index]
        dst = dst_doc.add_paragraph('', style=src.style)
        index = Word._find_paragraph(dst_doc, dst)
        # copy runs
        for run in src.runs:
            output_run = dst.add_run(run.text, style=run.style)
            output_run.bold = run.bold  # Run's bold data
            output_run.italic = run.italic  # Run's italic data
            output_run.underline = run.underline  # Run's underline data
            output_run.font.color.rgb = run.font.color.rgb  # Run's color data
        # copy paragraph_format
        src_fmt = copy.deepcopy(src.paragraph_format.element)
        dst_fmt = dst.paragraph_format.element
        dst_fmt.getparent().replace(dst_fmt, src_fmt)
        # move to new place if needed
        if new_index:
            return self.move_paragraph(index, new_index)
        else:
            return index

    def delete_paragraph(self, index):
        # noinspection PyProtectedMember
        p = self.paragraphs[index]._element
        p.getparent().remove_sheet(p)
        p._p = p._element = None

    def move_paragraph(self, old_index, new_index):
        if old_index == new_index and 0 < old_index < len(self.paragraphs):
            return old_index

        # noinspection PyProtectedMember
        p = self.paragraphs[old_index]._element
        parent = p.getparent()
        parent.remove_sheet(p)
        parent.insert(new_index, p)
        index = parent.index(p)
        return index

    def print(self, **kwargs):
        """ 使用打印机打印文档 """
        print("printing", self.filename, '...')

        app, document = Win32.open_word(self._absolute_filename, False)
        document.PrintOut(**kwargs)
        time.sleep(1)
        Win32.close(app, document)



import os
import sys
import subprocess
import time
import json
import io
from io import BytesIO
from pathlib import Path
from types import MethodType
from .base import BaseFile
from .util import Util
from .win32 import Win32


try:
    import PyPDF4
    from PyPDF4.pdf import PdfFileReader, PdfFileWriter
    from PyPDF4.generic import TextStringObject
except ImportError:
    PyPDF4 = None
    PdfFileReader = None
    PdfFileWriter = None
    TextStringObject = None


try:
    import reportlab
    from reportlab.pdfgen.canvas import Canvas
    from reportlab.pdfbase import pdfmetrics
    from reportlab.pdfbase.ttfonts import TTFont
except ImportError:
    reportlab = None
    Canvas = None
    pdfmetrics = None
    TTFont = None


try:
    import PIL
    from PIL import Image, ImageDraw, ImageEnhance
except ImportError:
    PIL = None
    Image = None
    ImageDraw = None
    ImageEnhance = None


class FontLib:
    ChineseFontNames = {
        '黑体': 'SimHei',
        '宋体': 'SimSunB',
        '仿宋': 'SimFang',
        '楷体': 'SimKai',
        '幼圆': 'SimYou',
        '隶书': 'SimLi',
    }
    #  OS Font paths
    MSFolders = r'Software\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders'

    MSFontDirectories = [
        r'SOFTWARE\Microsoft\Windows NT\CurrentVersion\Fonts',
        r'SOFTWARE\Microsoft\Windows\CurrentVersion\Fonts']

    X11FontDirectories = [
        # an old standard installation point
        "/usr/X11R6/lib/X11/fonts/TTF/",
        "/usr/X11/lib/X11/fonts",
        # here is the new standard location for fonts
        "/usr/share/fonts/",
        # documented as a good place to install new fonts
        "/usr/local/share/fonts/",
        # common application, not really useful
        "/usr/lib/openoffice/share/fonts/truetype/",
    ]

    OSXFontDirectories = [
        "/Library/Fonts/",
        "/Network/Library/Fonts/",
        "/System/Library/Fonts/",
        # fonts installed via MacPorts
        "/opt/local/share/fonts",
    ]

    # if not USE_FONTCONFIG and sys.platform != 'win32':
    #     OSXFontDirectories.append(str(Path.home() / "Library/Fonts"))
    #     X11FontDirectories.append(str(Path.home() / ".fonts"))

    @staticmethod
    def win32FontDirectory():
        r"""
        Return the user-specified font directory for Win32.  This is
        looked up from the registry key::

          \\HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders\Fonts

        If the key is not found, $WINDIR/Fonts will be returned.
        """
        import winreg
        try:
            with winreg.OpenKey(winreg.HKEY_CURRENT_USER, FontLib.MSFolders) as user:
                return winreg.QueryValueEx(user, 'Fonts')[0]
        except OSError:
            return os.path.join(os.environ['WINDIR'], 'Fonts')

    @staticmethod
    def get_fontext_synonyms(font_ext):
        """
        Return a list of file extensions that are synonyms for
        the given file extension *file_ext*.
        """
        return {'ttf': ('ttf', 'otf'),
                'otf': ('ttf', 'otf'),
                'afm': ('afm',)}[font_ext]

    @staticmethod
    def list_fonts(directory, extensions):
        """
        Return a list of all fonts matching any of the extensions, found
        recursively under the directory.
        """
        extensions = ["." + ext for ext in extensions]
        return [os.path.join(dirpath, filename)
                # os.walk ignores access errors, unlike Path.glob.
                for dirpath, _, filenames in os.walk(directory)
                for filename in filenames
                if Path(filename).suffix.lower() in extensions]

    @staticmethod
    def win32InstalledFonts(directory=None, font_ext='ttf'):
        """
        Search for fonts in the specified font directory, or use the
        system directories if none given.  A list of TrueType font
        filenames are returned by default, or AFM fonts if *font_ext* ==
        'afm'.
        """
        import winreg

        if directory is None:
            directory = FontLib.win32FontDirectory()

        font_ext = ['.' + ext for ext in FontLib.get_fontext_synonyms(font_ext)]

        items = set()
        for font_dir in FontLib.MSFontDirectories:
            try:
                with winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, font_dir) as local:
                    for j in range(winreg.QueryInfoKey(local)[1]):
                        key, direc, tp = winreg.EnumValue(local, j)
                        if not isinstance(direc, str):
                            continue
                        # Work around for https://bugs.python.org/issue25778,
                        # which is fixed in Py>=3.6.1.
                        direc = direc.split("\0", 1)[0]
                        try:
                            path = Path(directory, direc).resolve()
                        except (FileNotFoundError, RuntimeError):
                            # Don't fail with invalid entries (FileNotFoundError is
                            # only necessary on Py3.5).
                            continue
                        if path.suffix.lower() in font_ext:
                            items.add(str(path))
            except (OSError, MemoryError):
                continue
        return list(items)

    @staticmethod
    def OSXInstalledFonts(directories=None, font_ext='ttf'):
        """Get list of font files on OS X."""
        if directories is None:
            directories = FontLib.OSXFontDirectories
        return [path
                for directory in directories
                for path in FontLib.list_fonts(directory, FontLib.get_fontext_synonyms(font_ext))]

    @staticmethod
    def _call_fc_list():
        """Cache and list the font filenames known to `fc-list`.
        """
        # # Delay the warning by 5s.
        # timer = Timer(5, lambda: warnings.warn(
        #     'Matplotlib is building the font cache using fc-list. '
        #     'This may take a moment.'))
        # timer.start()
        # try:
        #     out = subprocess.check_output(['fc-list', '--format=%{file}\\n'])
        # except (OSError, subprocess.CalledProcessError):
        #     return []
        # finally:
        #     timer.cancel()
        # return [os.fsdecode(filename) for filename in out.split(b'\n')]
        return []

    @staticmethod
    def get_fontconfig_fonts(font_ext='ttf'):
        """List the font filenames known to `fc-list` having the given extension.
        """
        font_ext = ['.' + ext for ext in FontLib.get_fontext_synonyms(font_ext)]
        return [fname for fname in FontLib._call_fc_list()
                if Path(fname).suffix.lower() in font_ext]

    @staticmethod
    def findSystemFonts(font_paths=None, font_ext='ttf'):
        """
        Search for fonts in the specified font paths.  If no paths are
        given, will use a standard set of system paths, as well as the
        list of fonts tracked by fontconfig if fontconfig is installed and
        available.  A list of TrueType fonts are returned by default with
        AFM fonts as an option.
        """
        font_files = set()
        font_exts = FontLib.get_fontext_synonyms(font_ext)

        if font_paths is None:
            if sys.platform == 'win32':
                font_paths = [FontLib.win32FontDirectory()]
                # now get all installed fonts directly...
                font_files.update(FontLib.win32InstalledFonts(font_ext=font_ext))
            else:
                font_paths = FontLib.X11FontDirectories
                font_files.update(FontLib.get_fontconfig_fonts(font_ext))
                # check for OS X & load its fonts if present
                if sys.platform == 'darwin':
                    font_files.update(FontLib.OSXInstalledFonts(font_ext=font_ext))

        elif isinstance(font_paths, str):
            font_paths = [font_paths]

        for path in font_paths:
            font_files.update(map(os.path.abspath, FontLib.list_fonts(path, font_exts)))

        return [fname for fname in font_files if os.path.exists(fname)]

    @staticmethod
    def replaceFontName(font_name):
        """ replace font name """
        if font_name in FontLib.ChineseFontNames:
            return FontLib.ChineseFontNames[font_name]
        return font_name

    @staticmethod
    def findFontFile(font_name):
        """ Find font file"""
        font_files = FontLib.findSystemFonts()
        target = (FontLib.replaceFontName(font_name) + '.ttf').lower()
        for f in font_files:
            if f.lower().find(target) >= 0:
                return f


# PPT .pdf 文档类
class PDF(BaseFile):
    FONT_NAME = 'Helvetica-Bold'  # 西文字体(水印)
    CHINESE_FONT_NAME = 'SimHei'  # 中文字体(水印)

    class Pages:
        """ 页面集合类 """
        def __init__(self, pdf):
            self.pdf = pdf

        def __len__(self):
            return self.pdf.page_count

        def _add_page_method(self, page, index):
            """ add method to page object """
            def func_text(this, layout=False):
                # noinspection PyProtectedMember
                return this.pdf_obj._page_text(this.index, layout)

            def func_tables(this):
                # noinspection PyProtectedMember
                return this.pdf_obj._page_tables(this.index)

            # https://www.ianlewis.org/en/dynamically-adding-method-classes-or-class-instanc
            page.index = index
            page.pdf_obj = self.pdf
            page.text = MethodType(func_text, page)
            page.tables = MethodType(func_tables, page)
            return page

        def __getitem__(self, item):
            if 0 <= item < self.pdf.page_count:
                # noinspection PyProtectedMember
                page = self.pdf._obj.getPage(item)
                return self._add_page_method(page, item)
            raise IndexError('Index %s out of bound' % repr(item))

    def __init__(self, filename, template=None, **kwargs):
        self._watermark = None  # 水印pdf
        self._info = None  # pdf 文件信息对象
        self._register_fonts = {}  # 已注册的字体
        self._file_obj = None
        super().__init__(filename, template, **kwargs)

    def _open(self, filename, **kwargs):
        """  kwargs support:  password='xxx' """
        password = kwargs.get('password', None)

        f = open(filename, 'rb')
        self._file_obj = io.BytesIO(f.read())
        f.close()

        self._obj = PdfFileReader(self._file_obj, strict=False)
        self.filename = filename

        if password and self.is_encrypted:
            self._obj.decrypt(password=password)
        return True

    def _close(self, **kwargs):
        self._obj = None
        self._file_obj = None
        return True

    @staticmethod
    def _open_data_file(filename):
        with open(filename, 'rb') as f:
            f.read()

    @staticmethod
    def _save_data_to_file(filename, data, encoding='utf-8'):
        """ 将data存入文件"""
        file = open(filename, "wb")
        if isinstance(data, PdfFileWriter):
            data.write(file)
        elif isinstance(data, BytesIO):
            file.write(data.getbuffer())
        elif isinstance(data, bytes):
            file.write(data)
        elif isinstance(data, str):
            file.write(data.encode(encoding))
        elif isinstance(data, dict) or isinstance(data, list):
            file.write(json.dumps(data).encode(encoding))
        else:
            file.write(data)
        file.close()

    def _save_file(self, filename, **kwargs):
        file_ext = Util.file_ext(filename)
        if file_ext == '.pdf':
            return self._save_pdf(filename, **kwargs)
        else:
            return super()._save_file(filename, **kwargs)

    def _save_pdf(self, filename, **kwargs):
        """
        保存为pdf文件

        :param filename:  pdf文件名
        :param  kwargs:   support:  password='xxx'
        :return: 成功返回True, 失败返回False
        """
        # process parameters
        password = kwargs.get('password', None)
        watermark = kwargs.get('watermark', None)

        if watermark:
            self.watermark(str(watermark))

        pages = kwargs.get('pages', None)
        if isinstance(pages, int):
            pages = [pages]

        filename = self.filename if filename is None else filename

        wm_reader = None if self._watermark is None else PdfFileReader(self._watermark)

        # copy pages
        pdf_writer = PdfFileWriter()
        for idx, page in enumerate(range(self._obj.getNumPages())):
            # filter by pages
            if pages and idx not in pages:
                continue
            page = self._obj.getPage(page)
            # merge watermark
            if wm_reader:
                p = wm_reader.getPage(0)
                page.mergePage(p)
                pdf_writer.addPage(page)
            else:
                pdf_writer.addPage(page)

        # encrypt if needed
        if password:
            pdf_writer.encrypt(user_pwd=password, owner_pwd=None, use_128bit=True)

        self._save_data_to_file(filename, pdf_writer)
        return True

    @property
    def is_encrypted(self):
        """ pdf是否已加密 """
        return self._obj.isEncrypted

    @property
    def info(self) -> PyPDF4.pdf.DocumentInformation:
        """ 返回 pdf 文件信息对象"""
        if self._info is None:
            if not self.is_encrypted:
                self._info = self._obj.getDocumentInfo()
        return self._info

    @property
    def page_count(self):
        """ 返回页面数量 """
        return self._obj.getNumPages()

    def __len__(self):
        return self.page_count

    @property
    def pages(self):
        """ 返回页面的集合 """
        return PDF.Pages(self)

    def _register_font(self, font_name):
        """ 注册字体 """
        if font_name in self._register_fonts:
            return True

        # https://blog.csdn.net/a_faint_hope/article/details/90084992
        font_file = FontLib.findFontFile(font_name)
        if not font_file:
            raise ValueError('Cannot find font %s' % repr(font_name))
        pdfmetrics.registerFont(TTFont(font_name, font_file))
        self._register_fonts[font_name] = True

    def _get_font_name(self, text: str):
        """ 取得字体名称 """
        if Util.has_chinese(text):
            self._register_font(PDF.CHINESE_FONT_NAME)
            return PDF.CHINESE_FONT_NAME
        else:
            return PDF.FONT_NAME

    def watermark(self, text: str, color=(190, 190, 190), rotate=15, font_size=40):
        """
        加水印文字

        :param text:  水印文字
        :param color: (可选) 颜色, 数据类型是 tuple (r, g, b), 默认值是(190, 190, 190)
        :param rotate: (可选) 文字旋转角度
        :param font_size: (可选) 文字大小, 默认值是40
        :return: self
        """
        if not text:
            self._watermark = None
            return self

        if self.page_count > 0:
            page = self._obj.getPage(0)
            width = page.mediaBox.getWidth()
            height = page.mediaBox.getHeight()
        else:
            # default size is A4
            width = 596
            height = 842

        # https://www.thepythoncode.com/article/watermark-in-pdf-using-python
        # Generate the output to a memory buffer
        output_buffer = BytesIO()
        # Set Page Size
        c = Canvas(output_buffer, pagesize=(width, height))

        # Set the size and type of the font
        c.setFont(self._get_font_name(text), font_size)
        # add image instead of text: c.drawImage("logo.png", X, Y, 160, 160)

        # Set the color
        if isinstance(color, tuple):
            ct = (r / 255 for r in color)
            c.setFillColorRGB(*ct)
        else:
            c.setFillColor(color)

        # set rotation angle in degree
        c.rotate(rotate)

        # draw multiple text
        size = font_size
        for x in range(size, width, (len(text) * 2) * size):
            for y in range(-2 * size, height, 4 * size):
                c.drawString(x, y, text)

        c.save()
        self._watermark = output_buffer
        return self

    def _watermark_real_text(self, wm_text: str):
        # Generate a watermark pdf in output_buffer
        prefix = '_@#%^$_'
        output_buffer = BytesIO()
        c = Canvas(output_buffer, pagesize=(596, 842))
        c.setFont(self._get_font_name(wm_text), 40)
        c.drawString(10, 10, prefix + wm_text)
        c.save()

        # find the watermark text in the pdf
        wm_reader = PdfFileReader(output_buffer)
        page = wm_reader.getPage(0)
        content_object = page["/Contents"].getObject()
        content = PyPDF4.pdf.ContentStream(content_object, wm_reader)
        for operands, operator in content.operations:
            if operator == PyPDF4.utils.b_("Tj"):
                text = operands[0]
                if isinstance(text, str) and text.startswith(prefix):
                    text = text[len(prefix):]
                    return text
        return wm_text

    def _unwatermark(self, save_filename, watermark_text: str, password: str = None):
        """
        删除水印文字

        :param save_filename:  删除水印文字后的存盘 pdf 文件名
        :param watermark_text:  水印文字
        :param password:  (可选)密码
        :return: self
        """
        real_text = self._watermark_real_text(watermark_text)
        pdf_writer = PdfFileWriter()
        # for each page
        for p in range(self._obj.getNumPages()):
            page = self._obj.getPage(p)
            # Get the page content
            content_object = page["/Contents"].getObject()
            content = PyPDF4.pdf.ContentStream(content_object, self._obj)
            # Loop through all the page elements
            for operands, operator in content.operations:
                # Checks the TJ operator and replaces the corresponding string operand (Watermark text) with ''
                if operator == PyPDF4.utils.b_("Tj"):
                    text = operands[0]
                    if isinstance(text, str) and (text == watermark_text or text == real_text):
                        operands[0] = TextStringObject('')
            page.__setitem__(PyPDF4.generic.NameObject('/Contents'), content)
            pdf_writer.addPage(page)
        if password:
            pdf_writer.encrypt(user_pwd=password, owner_pwd=None, use_128bit=True)
        self._save_data_to_file(save_filename, pdf_writer)
        return self

    def _page_text(self, page_index, layout=False):
        """
        提取 指定页面中 的文字
        @return:
        """
        pdfplumber = Util.module('pdfplumber', True)

        # https://blog.csdn.net/h123456789999999/article/details/122832328
        with pdfplumber.open(self.filename) as pdf:
            if isinstance(page_index, int):
                page = pdf.pages[page_index]
                return page.extract_text(layout=layout)
            else:
                result = ''
                for page in pdf.pages:
                    result += page.extract_text(layout=layout)
                return result

    def _page_tables(self, page_index):
        """
        提取PDF中的图表数据
        @return:
        """
        pdfplumber = Util.module('pdfplumber', True)

        # https://blog.csdn.net/h123456789999999/article/details/122832328
        with pdfplumber.open(self.filename) as pdf:
            if isinstance(page_index, int):
                page = pdf.pages[page_index]
                tables = page.extract_tables()
                return tables
            else:
                result = []
                for page in pdf.pages:
                    t = page.extract_tables()
                    if isinstance(t, list):
                        result += t
                return result

    def text(self, layout=False):
        """ 返回 pdf 所有页的文本 """
        return self._page_text("all")

    def tables(self):
        """ 返回 pdf 所有页的表格数据 """
        return self._page_tables("all")

    def guess_password(self, words_filename: str = None):
        """
        (猜测) 破解 pdf 文件密码

        :param words_filename: (可选)密码文件名
        :return: 成功则返回密码，密码为空时返回空字符串。失败则返回None
        """
        def test(words, counter):
            for w in words:
                counter += 1
                if counter > 0 and counter % 2000 == 0:
                    print('guessed passwords %s ...' % counter)
                if self._obj.decrypt(password=w):
                    return w, counter
            return None, counter

        if not self.is_encrypted:
            return ''

        counter = 0

        if words_filename:
            # load passwords from file line by line
            f = open(words_filename, "r")
            line = f.readline()
            while line:
                word = line.strip()
                if self._obj.decrypt(password=word):
                    f.close()
                    return word
                line = f.readline()
            f.close()
        else:
            # get common passwords
            passwords = Util.common_passwords()
            password, counter = test(passwords, counter)
            if password is not None:
                return password

    def print(self, **kwargs):
        """ 使用打印机打印文档 """

        def com_object_print():
            """ print PDF by COM object """
            # noinspection PyBroadException
            try:
                # https://community.spiceworks.com/topic/2146606-automatically-printing-pdf-files-re-sizing-print-double-sided
                # http://www.verycomputer.com/340_bbd9c02270bc1c17_1.htm
                app = Win32.create_com_object('AcroExch.App')
                doc = Win32.create_com_object("AcroExch.AVDoc")
                doc.Open(self.filename, "")
                # Get the number of pages for this pdf, Subtract one because the count is 0 based
                max_page = doc.GetPDDoc().GetNumPages() - 1
                # Print Entire Document, Last value is ShrinkToFit
                shrink_to_fit = True
                postscript_level = 2
                doc.PrintPages(0, max_page, postscript_level, True, shrink_to_fit)
                # Print first two pages : doc.PrintPages(0, 1, 2, 1, 1)
                # quit
                app.Exit()
                doc.Close(True)
                return True
            except Exception:
                return False

        def find_acrobat_exe():
            """ find Acrobat executable filename """
            # noinspection PyBroadException
            try:
                import winreg
                exe = winreg.QueryValue(winreg.HKEY_CLASSES_ROOT, 'Software\\Adobe\\Acrobat\\Exe')
                return exe
            except Exception:
                return None

        def command_line_print():
            """ print PDF by command line """
            # https://zgserver.com/pdf-26.html
            acrobat_exe = find_acrobat_exe()
            if not acrobat_exe or not os.path.exists(acrobat_exe):
                raise ValueError("Adobe Acrobat is needed to print .pdf file")

            # noinspection PyBroadException
            try:
                # compose command of acrobat print
                cmd = '"%s" /N /S /T "%s"' % (acrobat_exe, self._absolute_filename)
                proc = subprocess.Popen(cmd)
                # wait for print pdf to the printer.
                time.sleep(6)
                # kill process
                proc.kill()
                # os.system("TASKKILL /F /IM %s" % Util.file_name(acrobat_exe))
                return True
            except Exception:
                return False

        print("printing", self.filename, '...')

        if not com_object_print():
            return command_line_print()
        return True




import os
import shutil
from .util import Util
from .variable import VarData


class BaseFile:
    """
    Base class for Office files, include .docx, .xlsx, .pptx, .pdf.

    :Chinese: Office 文档基类, 包括：.docx, .xlsx, .pptx, .pdf.
    """
    def __init__(self, filename=None, template=None, **kwargs):
        """
        Init and open file depends on template file.

        :Chinese: 初始化，并打开文件

        :param filename:   file name
        :param template:  (optional) template file name. if template is provided,
            template file is copied to overwrite existing file or create new file.

            :Chinese:(可选)模板文件名。如果模板不为空，则复制模板文件创建文件或覆盖已有文件。
        :return: self
        """
        self.filename = None
        self._obj = None
        self.open(filename, template, **kwargs)

    def open(self, filename, template=None, **kwargs):
        """
        Open file.

        :Chinese: 打开文件

        :param filename:   filename.
        :param template:  (optional) template filename. If template is provided,
            template file is copied to overwrite existing file or create new file.

            :Chinese:(可选)模板文件名。如果模板不为空，则复制模板文件创建文件或覆盖已有文件。
        :return: self
        """
        self.close()

        if filename and template:
            if os.path.exists(template):
                shutil.copyfile(template, filename)
            else:
                raise FileNotFoundError('template file %s not found' % repr(template))

        if filename and os.path.exists(filename):
            self._open(filename, **kwargs)
        else:
            self._new_file(filename, **kwargs)

        return self

    def close(self, **kwargs):
        """
        Close file.

        :Chinese: 关闭文件.

        :return: self
        """
        if self._close(**kwargs):
            self._obj = None
        self.filename = None
        return self

    def save(self, filename=None, **kwargs):
        """
        Save file.

        :Chinese: 保存文件

        :param filename:  filename to save. if None, save to original file.
            :Chinese: (可选)存盘文件名。如果缺省，则保存到原文件名。
        :return: self
        """
        filename = self.filename if filename is None else filename
        self._save_file(filename, **kwargs)
        return self

    def _new_file(self, filename, **kwargs):
        """ Extend class should implement this method """
        raise IOError('new file is not supported')

    def _save_file(self, filename, **kwargs):
        """ Extend class should implement this method """
        raise IOError('save to %s is not supported' % repr(Util.file_ext(filename)))

    def _open(self, filename, **kwargs):
        """ Extend class should implement this method """
        raise IOError('open file is not supported')

    # noinspection PyMethodMayBeStatic
    # noinspection PyUnusedLocal
    def _close(self, **kwargs):
        """ Extend class should implement this method """
        return True

    def fill(self, data):
        """
        Fill data into the document, replace variable tags with its values.

        :Chinese: 将数据填入, 替换变量标签，生成文件内容。

        :param data:  the data to fill with, could be a dict,
                    or an Excel filename, or a DataFrame object.
            :Chinese: dictionary对象 或 excel文件名
        :return: self
        """
        data = VarData.open(data)

        # 填入数据
        self._fill_data(data)

        data.close()

        return self

    def _fill_data(self, data):
        """ Extend class should implement this method to fill data into the file"""
        raise IOError('fill data is not implemented')

    @property
    def _absolute_filename(self):
        """ return file name with absolute path """
        filename = self.filename
        if not Util.is_absolute_path(filename):
            filename = os.path.join(os.getcwd(), filename)
        return filename

    # noinspection PyUnusedLocal
    def print(self, **kwargs):
        """
        Print the file.

        Extend class should implement this method.
        """
        pass
